import os

from scene_engine.rest.helpers.annotations import coco_annotation
from shapely.geometry import Polygon
from shared.coordinate_interaction_enricher.enricher import (
    coordinate_interaction_enricher, create_polygon_list)
from shared.external.constants import AnnotationTypes
from shared.external.custom_exceptions import InvalidAnnotationError
from shared.external.models.annotation import AnnotationEntity
from shared.external.models.annotation_transformers.scale import \
    ScaleAnnotation
from shared.external.tools.logger import get_logger
from shared.tools.misc import json_file_to_dic

logger = get_logger(__name__)

my_path = path = os.path.dirname(os.path.abspath(__file__))


class TestCoco:
    def test_coco(self):
        """Simple unit test to test the coco annotation generator."""

        annotation_metadata = [
            {
                "id": "ooc_1.000.0000000003-cor.ann",
                "session_name": "ooc_1",
                "timestamp": "2020-07-31T19:13:31.004260+00:00",
                "asset_id": "ooc_1.000.0000000003-cor.png",
                "media_type": "images",
                "provider": "rendered",
                "annotation_type": "two_d_bounding_box",
                "annotation_entities": [],
                "sets": ["2ef58f20-cd29-437d-aac5-b71f2d267431"],
                "binary_id": None,
                "labels": [],
            },
            {
                "id": "ooc_1.000.0000000010-cor.ann",
                "session_name": "ooc_1",
                "timestamp": "2020-07-31T19:14:06.435951+00:00",
                "asset_id": "ooc_1.000.0000000010-cor.png",
                "media_type": "images",
                "provider": "rendered",
                "annotation_type": "two_d_bounding_box",
                "annotation_entities": [
                    {
                        "uid": "ec868ba6f16a8a2b4a5622af2b76c733",
                        "label": "car",
                        "annotation_type": "two_d_bounding_box",
                        "attributes": None,
                        "url": None,
                        "resource": None,
                        "coordinates": [
                            {"x": 167.5, "y": 121.5},
                            {"x": 194.5, "y": 121.5},
                            {"x": 194.5, "y": 157.5},
                            {"x": 167.5, "y": 157.5},
                        ],
                        "related_annotations": [],
                        "binary_id": None,
                        "category_id": 1,
                    }
                ],
                "sets": ["2ef58f20-cd29-437d-aac5-b71f2d267431"],
                "binary_id": None,
                "labels": ["car"],
            },
        ]
        images_metadata = [
            {
                "session_name": "ooc_1",
                "timestamp": "2020-03-20T19:00:00+00:00",
                "filename": "ooc_1.000.0000000003-cor.png",
                "sets": ["67eb75f4-9d5a-45f5-b555-d8ef231007b4"],
                "width": 256,
                "height": 256,
                "background": "travis_afb_001",
                "thumbnails": {
                    "tiny": {
                        "resolution": {"height": 100, "width": 100},
                        "filename": "ooc_1_tiny.000.0000000003-cor.png"
                    },
                    "small": {
                        "resolution": {"height": 300, "width": 300},
                        "filename": "ooc_1_small.000.0000000003-cor.png"
                    }
                }
            },
            {
                "session_name": "ooc_1",
                "timestamp": "2020-03-20T19:00:00+00:00",
                "filename": "ooc_1.000.0000000010-cor.png",
                "sets": ["67eb75f4-9d5a-45f5-b555-d8ef231007b4"],
                "width": 256,
                "height": 256,
                "background": "travis_afb_001",
                "thumbnails": {
                    "tiny": {
                        "resolution": {"height": 100, "width": 100},
                        "filename": "ooc_1_tiny.000.0000000010-cor.png"
                    },
                    "small": {
                        "resolution": {"height": 300, "width": 300},
                        "filename": "ooc_1_small.000.0000000010-cor.png"
                    }
                }
            }
        ]
        timestamp = "2020-04-04"
        year = "2020"
        description = "Mock annotation set"
        res = coco_annotation(
            annotation_metadata, images_metadata, timestamp, year, description
        )
        expected_res = {
            "info": {
                "description": "Mock annotation set",
                "url": "",
                "version": "",
                "year": "2020",
                "contributor": "",
                "date_created": "2020-04-04",
            },
            "licenses": [],
            "images": [
                {
                    "license": "",
                    "file_name": "ooc_1.000.0000000003-cor.png",
                    "coco_url": "",
                    "height": 256,
                    "width": 256,
                    "date_captured": "2020-03-20T19:00:00+00:00",
                    "flickr_url": "",
                    "id": 0,
                },
                {
                    "license": "",
                    "file_name": "ooc_1.000.0000000010-cor.png",
                    "coco_url": "",
                    "height": 256,
                    "width": 256,
                    "date_captured": "2020-03-20T19:00:00+00:00",
                    "flickr_url": "",
                    "id": 1,
                },
            ],
            "annotations": [
                {
                    "iscrowd": 0,
                    "image_id": 1,
                    "id": 0,
                    "category_id": 1,
                    "bbox": [167.5, 121.5, 27.0, 36.0],
                }
            ],
            "categories": [{"supercategory": "", "name": "car", "id": 1}],
        }

        assert res == expected_res


class TestAnnotationEntity:
    def test_creation(self):
        coordinates = [{'x': 448.0, 'y': 467.0},
                       {'x': 485.0, 'y': 467.0},
                       {'x': 485.0, 'y': 487.0},
                       {'x': 448.0, 'y': 487.0}]
        confidence = 0.75
        annotation_entity = AnnotationEntity(
            label="car",
            annotation_type=AnnotationTypes.TWO_D_BOUNDING_BOX,
            coordinates=coordinates,
            confidence=confidence
        )
        assert annotation_entity.label == "car"
        assert annotation_entity.confidence == 0.75

    def test_error_at_creation(self):
        try:
            coordinates = [{'x': 448.0, 'y': 467.0},
                           {'x': 485.0, 'y': 467.0},
                           {'x': 485.0, 'y': 487.0},
                           {'x': 448.0, 'y': 487.0}]
            confidence = 1
            annotation_entity = AnnotationEntity(
                label="car",
                annotation_type=AnnotationTypes.TWO_D_BOUNDING_BOX,
                coordinates=coordinates,
                confidence=confidence
            )
        except InvalidAnnotationError as e:
            assert "invalid confidence format <class 'int'>" == e.__str__()


class TestAnnotationCoordinateInteraction:
    def test_polygon_creation(self):
        coordinates = [{'x': 448.0, 'y': 467.0},
                       {'x': 485.0, 'y': 467.0},
                       {'x': 485.0, 'y': 487.0},
                       {'x': 448.0, 'y': 487.0}]
        confidence = 0.75
        annotation_entities = [
            {
                "label": 'car',
                "annotation_type": AnnotationTypes.TWO_D_BOUNDING_BOX,
                "coordinates": coordinates,
                "confidence": confidence
            },
            {
                "label": 'car',
                "annotation_type": AnnotationTypes.TWO_D_BOUNDING_BOX,
                "coordinates": coordinates,
                "confidence": confidence
            }]
        polygons = create_polygon_list(
            entities=annotation_entities, label="car")
        assert len(polygons) == 2
        assert isinstance(polygons[0], Polygon)
        assert isinstance(polygons[1], Polygon)

    def test_annotation_coordinate_interaction(self):
        annotation_with_collision = ScaleAnnotation(json_file_to_dic(
            os.path.join(my_path, 'resources/annotation_interaction_with_collision.ann')))
        annotation_without_collision = ScaleAnnotation(json_file_to_dic(os.path.join(
            my_path, 'resources/annotation_interaction_without_collision.ann')))

        # 3 annotations entities, 2 trucks and 1 car.
        annotation_with_collision_dict = annotation_with_collision.to_dic()
        annotation_without_collision_dict = annotation_without_collision.to_dic()
        time_tolerance = 1
        assert (
            coordinate_interaction_enricher(
                annotation_meta=annotation_with_collision_dict,
                labels=[
                    "truck",
                    "car"])) == [
            annotation_with_collision.timestamp,
            annotation_with_collision.timestamp]
        assert (coordinate_interaction_enricher
                (annotation_meta=annotation_without_collision_dict,
                 labels=["truck", "car"])) == []
        #  Checks if two trucks are colliding which they are not
        #  with the given coordinates although car and truck still collide
        assert (coordinate_interaction_enricher
                (annotation_meta=annotation_with_collision_dict,
                 labels=["truck", "truck"])) == []
        # Checks if two cars are colliding but there is only one car in the
        # annotation
        assert (coordinate_interaction_enricher
                (annotation_meta=annotation_with_collision_dict,
                 labels=["car", "car"])) == []

    def test_annotation_coordinate_interaction_static_mask(self):
        annotation_with_collision = ScaleAnnotation(json_file_to_dic(
            os.path.join(my_path, 'resources/annotation_interaction_with_collision.ann')))

        static_masks = [{"label": "test_mask_1",
                         "coordinates": [{'x': 40.0, 'y': 25.0},
                                         {'x': 40.0, 'y': 10.0},
                                         {'x': 50.0, 'y': 10.0},
                                         {'x': 50.0, 'y': 25.0}]},
                        {"label": "test_mask_2",
                         "coordinates": [{'x': 30.0, 'y': 25.0},
                                         {'x': 30.0, 'y': 10.0},
                                         {'x': 40.0, 'y': 10.0},
                                         {'x': 40.0, 'y': 25.0}]},
                        {"label": "test_mask_3",
                         "coordinates": [{'x': 100.0, 'y': 25.0},
                                         {'x': 100.0, 'y': 10.0},
                                         {'x': 120.0, 'y': 10.0},
                                         {'x': 120.0, 'y': 25.0}]}]

        annotation_with_collision_dict = annotation_with_collision.to_dic()
        time_tolerance = 1
        assert (coordinate_interaction_enricher
                (annotation_meta=annotation_with_collision_dict,
                 labels=["test_mask_1", "car"],
                 static_masks=static_masks)) == []
        assert (coordinate_interaction_enricher
                (annotation_meta=annotation_with_collision_dict,
                 labels=["test_mask_3", "car"],
                 static_masks=static_masks)) == []
        assert (coordinate_interaction_enricher
                (annotation_meta=annotation_with_collision_dict,
                 labels=["test_mask_1", "truck"],
                 static_masks=static_masks)) == [annotation_with_collision.timestamp,
                                                 annotation_with_collision.timestamp]



from concurrent.futures.thread import ThreadPoolExecutor

from shared.external.tools.threading_utils import create_exception_map


def test_create_exception_map():
    def threaded_process(i):
        if i == 4:
            i / 0.0
        elif i == 40:
            raise TypeError
        else:
            return None

    nums = [i for i in range(100)]
    with ThreadPoolExecutor(max_workers=1) as executor:
        future_res = {
            executor.submit(
                threaded_process,
                i=num): num for num in nums}
        id_exception_map = create_exception_map(future_map=future_res)

    assert isinstance(id_exception_map[4], ZeroDivisionError)
    assert isinstance(id_exception_map[40], TypeError)



#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.
#
from typing import Dict, Optional

from ..tools import misc
from ..tools.naming import UriParser


class ObjectAccessError(Exception):
    pass


class ObjectAccessMedium:
    PUBLIC_URL = "public-url"
    GS = "gs"
    S3 = "s3"

# examples
# {
#   "medium": "public-url", #optional
#   "url": "http....", #required
#   "filename": "abc.png" #optional
# }
#
# {
#   "medium": "s3",  #optional
#   "uri": "s3://bucket/key", #required
#   "filename": "abc.png", #optional
# }
#


class ObjectAccess(object):
    def __init__(self,
                 filename: Optional[str] = None,
                 url: Optional[str] = None,
                 uri: Optional[str] = None):
        if url:
            self.medium = ObjectAccessMedium.PUBLIC_URL
        elif uri:
            uri_parser = UriParser(uri)
            if uri_parser.cloud_provider == "gs":
                self.medium = ObjectAccessMedium.GS
            elif uri_parser.cloud_provider == "s3":
                self.medium = ObjectAccessMedium.S3
            else:
                raise ObjectAccessError(
                    "invalid cloud storage provider {}".format(
                        uri_parser.cloud_provider))
            filename = uri_parser.key.split("/")[-1]
        else:
            raise ObjectAccessError(
                "invalid medium- either url, or uri should be specified")

        self.url = url  # optional
        self.uri = uri  # optional
        self.filename = filename  # optional

    def __hash__(self):
        if self.url:
            return self.url.__hash__()
        else:
            return self.uri.__hash__()

    def __eq__(self, other):
        if self.url and other.url:
            return self.url == other.url
        else:
            return self.uri == other.uri

    def to_dic(self) -> dict:
        object_access_dict = {}

        if self.filename:
            object_access_dict["filename"] = self.filename

        if self.url:
            object_access_dict["url"] = self.url

        if self.uri:
            object_access_dict["uri"] = self.uri

        return object_access_dict

    @classmethod
    def from_dict(cls, data: Dict):
        fields = {
            "filename",
            "url",
            "uri"
        }
        return cls(**{field: data.get(field) for field in fields})



"""Miscellaneous (catch all) tools Copyright 2020 Caliber Data Labs."""

#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.
#
from typing import NamedTuple


class VideoProperties(NamedTuple):
    width: int
    height: int
    fps: float
    duration: float
    num_frames: int
    num_read_frames: int


def get_video_attributes(video_uri: str) -> VideoProperties:
    import ffmpeg

    info = ffmpeg.probe(video_uri, count_frames=None)
    info_stream = None
    for streams in info["streams"]:
        if streams.get("codec_type") == "video":
            info_stream = streams
            break

    if not info_stream:
        info_stream = info['streams'][0]

    return VideoProperties(
        width=int(info_stream.get('width') or 0),
        height=int(info_stream.get('height') or 0),
        duration=float(info_stream['duration']),
        num_frames=int(info_stream.get('nb_frames') or 0),
        fps=0 if info_stream['avg_frame_rate'] == '0/0' else eval(info_stream['avg_frame_rate']),
        num_read_frames=int(info_stream['nb_read_frames'] or 0)
    )



#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.
#
import logging
import os
import sys
import time
import traceback

# LOGGING_LEVEL = logging.INFO
LOGGING_LEVEL = logging.DEBUG
# LOGGING_LEVEL = logging.ERROR


def get_logger(module_name):
    """Logger to stdout.

    Typical use: logger = get_module_logger(__name__)
    """
    logger = logging.getLogger(module_name)
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter(
        "%(asctime)s [%(name)s] %(levelname)s %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(LOGGING_LEVEL)
    return logger


def log_with_exc_info(logger, e):
    """Log with exception info.

    Common usage when logging at the end of a try/except to identify the
    line number and error type
    """
    # Log the message with the exception info
    logger.error("ERROR::: {}, {}".format(e, get_exception_info()))


def get_exception_info():
    """Get information on a raised exception."""
    exc_type, exc_obj, exc_tb = sys.exc_info()
    fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
    return {
        "traceback": repr(traceback.extract_tb(exc_tb)),
        "exception_type": exc_type,
        "file_name": fname,
        "line_number": exc_tb.tb_lineno,
    }


class DurationLogger:
    def __init__(self, tag: str = ''):
        self.tag = tag
        self.start_time = None
        self.end_time = None

    def start(self):
        self.start_time = time.time()

    def end(self):
        self.end_time = time.time()

    def duration(self, digs=2):
        return round(self.end_time - self.start_time, digs)

    def log_duration(self, tag=''):
        logging.info(
            "{} {} ::: Duration: = {}s".format(
                tag,
                self.tag,
                self.duration()))



#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.
#
import io
from datetime import datetime
from typing import Dict, List, NamedTuple, Optional, Tuple

from ..constants import (AnnotationMediaTypes, AnnotationTypes, AssetsConstants, AnnotationGroups)
from ..custom_exceptions import InvalidAnnotationError
from ..tools import misc, time_utils
from ..tools.logger import get_logger
from .object_access import ObjectAccess
from ..tools.misc import get_md5_from_string

logger = get_logger(__name__)


class Point(NamedTuple):
    left: float
    top: float
    label: str
    category_id: Optional[int] = None
    confidence: Optional[float] = None
    geometry: Optional[str] = None
    attributes: Optional[dict] = None


class BoundingBox(NamedTuple):
    x_min: float
    x_max: float
    y_min: float
    y_max: float
    label: str
    category_id: Optional[int] = None
    confidence: Optional[float] = None
    attributes: Optional[dict] = None


class Polygon(NamedTuple):
    coordinates: List[Tuple[float, float]]
    label: str
    category_id: Optional[int] = None
    confidence: Optional[float] = None
    attributes: Optional[dict] = None


class Line(NamedTuple):
    coordinates: List[Tuple[float, float]]
    label: str
    complete: Optional[bool] = False
    category_id: Optional[int] = None
    confidence: Optional[float] = None
    attributes: Optional[dict] = None


class Label(NamedTuple):
    label: str
    class_name: Optional[str] = None
    category_id: Optional[int] = None
    confidence: Optional[float] = None


class Pose(NamedTuple):
    label: str
    keypoints: List[Point]
    keypoint_names: List[str]
    attributes: Optional[dict] = None


class Annotation(object):
    def __init__(self, **kwargs):
        self.id: str = kwargs.get("id", misc.get_guid())
        self.timestamp: datetime = kwargs.get("timestamp", datetime.utcnow())

        self.asset_id: str = kwargs.get("asset_id") or kwargs.get("image_id") or \
                             kwargs.get("video_id") or kwargs.get("lidar_id")
        if not self.asset_id:
            raise InvalidAnnotationError("asset_id (image, video, lidar) should be specified")

        self.media_type: str = kwargs.get("media_type", AssetsConstants.IMAGES_ASSET_ID)
        if self.media_type not in AnnotationMediaTypes.VALID_TYPES:
            raise InvalidAnnotationError(
                "invalid annotation type {}".format(self.media_type)
            )

        self.annotation_type: str = kwargs.get("annotation_type")
        if self.annotation_type not in AnnotationTypes.VALID_TYPES:
            raise InvalidAnnotationError(
                "invalid annotation type {}".format(self.annotation_type)
            )

        self.provider: Optional[str] = kwargs.get("provider")
        self.annotation_entities: List[AnnotationEntity] = kwargs.get("annotation_entities", [])
        self.annotation_meta: Dict = kwargs.get("annotation_meta", {})
        self.sets: List[str] = kwargs.get("sets", [])
        if "set_id" in kwargs:
            self.sets.append(kwargs.get("set_id"))
        self.session_uid: Optional[str] = kwargs.get("session_uid")
        self.sensor: Optional[str] = kwargs.get("sensor")
        self.version: Optional[str] = kwargs.get("version")
        self.annotation_group: str = kwargs.get("annotation_group", AnnotationGroups.GROUND_TRUTH)
        self.masks: Dict[str, ObjectAccess] = kwargs.get("masks", {})
        self.tags: List[str] = kwargs.get("tags", [])
        self.frame: Optional[int] = kwargs.get("frame")
        if self.media_type in [AssetsConstants.VIDEOS_ASSET_ID] and self.frame is None:
            raise InvalidAnnotationError(
                "{} annotation need frame".format(self.media_type)
            )

    def set_asset_id(self, asset_id: str):
        self.asset_id = asset_id

    def finalize(self):
        self.set_annotation_entities()
        self.__sanity_check()

    def set_annotation_entities(self):
        pass

    def post_upload(self, annotations_amc, owner_organization_id: str):
        pass

    def __sanity_check(self):
        pass

    def to_dic(self):
        """
        return a dictionary object representative of Annotation
        :return: dictionary object
        """
        annotation_entities_ = []
        labels = set()
        for annotation_entity in self.annotation_entities:
            annotation_entities_.append(annotation_entity.to_dic())
            labels.add(annotation_entity.label)
        return {
            "session_uid": self.session_uid,
            "timestamp": time_utils.datetime_or_str_to_iso_utc(self.timestamp),
            "asset_id": self.asset_id,
            "media_type": self.media_type,
            "provider": self.provider,
            "annotation_type": self.annotation_type,
            "annotation_entities": annotation_entities_,
            "sets": self.sets,
            "labels": list(labels),
            "sensor": self.sensor,
            "version": self.version,
            "annotation_group": self.annotation_group,
            "masks": {k: v.to_dic() for k, v in self.masks.items()},
            "frame": self.frame,
            "tags": self.tags
        }


class RelatedAnnotation(object):
    # initialize a related annotation object(
    def __init__(self, related_annotation_dic):
        try:
            # annotation label
            self.uid = related_annotation_dic["uid"]

            # the unique (worldwide) id of the entity
            self.relationship = related_annotation_dic["relationship"]
            if self.relationship not in {"parent", "child", "sibling"}:
                raise InvalidAnnotationError(
                    "unrecognized relation {}".format(self.relationship)
                )

        except KeyError as e:
            raise InvalidAnnotationError(str(e))

    def to_dic(self):
        """
        return a dictionary object representative of Annotation entity
        :return: dictionary object
        """
        return {"uid": self.uid, "relationship": self.relationship}


class AnnotationEntity(object):
    # initialize an annotation object (a single annotation such as tire,
    # pedestrian, etc)
    def __init__(self,
                 label: str,
                 annotation_type: str,
                 class_name: Optional[str] = None,
                 coordinates: Optional[List] = None,
                 confidence: Optional[float] = None,
                 mask_color: Optional[str] = None,
                 mask_id: Optional[str] = None,
                 uid: Optional[str] = None,
                 category_id: Optional[int] = None,
                 related_annotations: Optional[List[RelatedAnnotation]] = None,
                 attributes: Optional[dict] = None,
                 mask_type: Optional[str] = None,
                 composition: Optional[List[Dict[str, Point]]] = None):

        # annotation label
        self.label = label
        self.category_id = category_id
        self.class_name = class_name
        if mask_color:
            self.mask_color = mask_color.lower()
            if not misc.is_valid_hex(self.mask_color):
                raise InvalidAnnotationError(
                    "invalid hex color {}".format(mask_color))
        else:
            self.mask_color = None

        # the unique (worldwide) id of the entity
        self.uid = uid or misc.get_guid()

        self.mask_id = mask_id

        if isinstance(confidence, float):
            self.confidence = confidence
            if self.confidence < 0 or self.confidence > 1:
                raise InvalidAnnotationError(
                    "confidence {} should be between 0 and 1".format(
                        self.confidence))
        elif confidence is None:
            self.confidence = None
        else:
            raise InvalidAnnotationError(
                "invalid confidence format {}".format(type(confidence))
            )

        if annotation_type not in AnnotationTypes.VALID_TYPES:
            raise InvalidAnnotationError(
                "invalid annotation type {}".format(annotation_type)
            )
        else:
            self.annotation_type = annotation_type

        # x, y coordinates
        self.coordinates = []

        for point in coordinates or []:
            self.coordinates.append(
                {"x": float(point["x"]), "y": float(point["y"])})

        # coordinate checks for various geometry types:
        if self.annotation_type == AnnotationTypes.POINT and len(
                self.coordinates) != 1:
            raise InvalidAnnotationError(
                "point annotations should exactly have 1 point"
            )

        if self.annotation_type == AnnotationTypes.TWO_D_BOUNDING_BOX and len(
                self.coordinates) != 4:
            raise InvalidAnnotationError(
                "bounding_box annotations should exactly have 4 points"
            )

        if self.annotation_type in {
                AnnotationTypes.POLYGON,
                AnnotationTypes.LINE} and not self.coordinates:
            raise InvalidAnnotationError(
                "empty coordinates for annotation_type = {}".format(
                    self.annotation_type))

        if self.annotation_type == AnnotationTypes.SEGMENTATION and not self.mask_color:
            raise InvalidAnnotationError(
                "mask color should be specified for segmentation"
            )

        # get all related_annotations
        self.related_annotations = []

        for related_annotation in related_annotations or []:
            self.related_annotations.append(
                RelatedAnnotation(related_annotation))

        # optional attributes of an entity
        self.attributes = attributes or {}

        self.mask_type = mask_type

        # composition makes sense for key point annotations only
        self.composition = composition or []
        if self.composition and self.annotation_type not in [AnnotationTypes.TWO_D_POSE]:
            raise InvalidAnnotationError("composition {} cannot be defined for {}".format(
                self.composition,
                self.annotation_type)
            )

    def to_dic(self):
        """
        return a dictionary object representative of Annotation entity
        :return: dictionary object
        """
        related_annotations_ = []
        for related_annotation in self.related_annotations:
            related_annotations_.append(related_annotation.to_dic())

        dic = {
            "uid": self.uid,
            "label": self.label,
            "annotation_type": self.annotation_type,
            "attributes": self.attributes,
            "related_annotations": related_annotations_,
            "category_id": self.category_id,
        }
        if self.coordinates:
            dic["coordinates"] = self.coordinates
        if self.confidence:
            dic["confidence"] = self.confidence
        if self.mask_id:
            dic["mask_id"] = self.mask_id
        if self.mask_color:
            dic["mask_color"] = self.mask_color
        if self.class_name:
            dic["class_name"] = self.class_name
        if self.mask_type:
            dic["mask_type"] = self.mask_type
        if self.composition:
            composition_dict = []
            for item in self.composition:
                composition_dict.append({key: point._asdict() for key, point in item.items()})
            dic["composition"] = composition_dict
        return dic

    @classmethod
    def from_dict(cls, annotation_entity_dic: dict):

        confidence = annotation_entity_dic.get("confidence")
        if isinstance(confidence, float):
            if confidence < 0 or confidence > 1:
                raise InvalidAnnotationError(
                    "confidence {} should be between 0 and 1".format(
                        confidence))
        elif confidence is None:
            pass
        else:
            raise InvalidAnnotationError(
                "invalid confidence format {}".format(type(confidence))
            )

        coordinates = []

        try:
            for point in annotation_entity_dic.get("coordinates", []):
                coordinates.append(
                    {"x": float(point["x"]), "y": float(point["y"])})

        except KeyError:
            raise InvalidAnnotationError(str(KeyError))

        related_annotations = []
        for related_annotation in annotation_entity_dic.get(
                "related_annotations", []):
            related_annotations.append(RelatedAnnotation(related_annotation))

        composition = annotation_entity_dic.get("composition", [])
        composition_points = []
        for item in composition:
            composition_points.append({key: Point(**point_dict) for key, point_dict in item.items()})

        return cls(
            label=annotation_entity_dic["label"],
            category_id=annotation_entity_dic.get(
                "category_id"),
            uid=annotation_entity_dic.get("uid", misc.get_guid()),
            annotation_type=annotation_entity_dic["annotation_type"],
            mask_id=annotation_entity_dic.get("mask_id"),
            confidence=confidence,
            mask_color=annotation_entity_dic.get("mask_color"),
            coordinates=coordinates,
            attributes=annotation_entity_dic.get("attributes"),
            class_name=annotation_entity_dic.get("class_name"),
            related_annotations=related_annotations,
            composition=composition_points
        )

# TODO replace image_id with asset_id in all annotation classes


class SegmentationAnnotation(Annotation):
    def __init__(self,
                 labels: List[str],
                 mask_colors: List[str],
                 mask_urls: Optional[List[str]] = None,
                 mask_uris: Optional[List[str]] = None,
                 mask_file_paths: Optional[List[str]] = None,
                 mask_ids: Optional[List[str]] = None,
                 mask_types: Optional[List[str]] = None,
                 confidences: Optional[List[float]] = None,
                 **kwargs
                 ):
        super().__init__(annotation_type=AnnotationTypes.SEGMENTATION, **kwargs)
        self.labels = labels
        self.mask_urls = mask_urls or [None] * len(labels)
        self.mask_uris = mask_uris or [None] * len(labels)
        self.mask_ids = mask_ids or [None] * len(labels)
        self.mask_file_paths = mask_file_paths or [None] * len(labels)
        self.mask_colors = mask_colors
        self.mask_types = mask_types or [None] * len(labels)
        self.confidences = confidences or [None] * len(labels)
        self.mask_id_path_map = {}
        self.finalize()

    def post_upload(self, annotations_amc, owner_organization_id: str):
        for mask_id, mask_file_path in self.mask_id_path_map.items():
            with open(mask_file_path, 'rb') as f:
                file_obj = io.BytesIO(f.read())
            object_access = annotations_amc.put_file(
                id=mask_id,
                file_object=file_obj,
                owner_organization_id=owner_organization_id,
                content_type="image/png"
            )
            self.masks[mask_id] = object_access

    def set_annotation_entities(self):
        self.annotation_entities = []
        assert len(self.labels) == len(self.mask_urls)
        assert len(self.labels) == len(self.confidences)
        assert len(self.labels) == len(self.mask_colors)
        assert len(self.labels) == len(self.mask_uris)
        assert len(self.labels) == len(self.mask_file_paths)
        assert len(self.labels) == len(self.mask_ids)
        assert len(self.labels) == len(self.mask_types)
        masks = {}
        for label, url, confidence, mask_color, uri, mask_id, mask_file_path, mask_type in zip(
                self.labels, self.mask_urls, self.confidences, self.mask_colors, self.mask_uris,
                self.mask_ids, self.mask_file_paths, self.mask_types):
            if url:
                mask_id = mask_id or str(get_md5_from_string(url)).replace('-', '_')
                object_access = ObjectAccess(url=url)
            elif uri:
                mask_id = mask_id or str(get_md5_from_string(uri)).replace('-', '_')
                object_access = ObjectAccess(uri=uri)
            elif mask_file_path:
                mask_id = mask_id or str(get_md5_from_string(mask_file_path)).replace('-', '_')
                self.mask_id_path_map[mask_id] = mask_file_path
                object_access = None
            else:
                raise InvalidAnnotationError(
                    "either url or uri or mask file path should be specified for {}".format(
                        self.id))
            if not mask_id:
                mask_id = misc.get_guid()
            masks[mask_id] = object_access
            self.annotation_entities.append(
                AnnotationEntity(
                    label=label,
                    annotation_type=AnnotationTypes.SEGMENTATION,
                    mask_id=mask_id,
                    confidence=confidence,
                    mask_color=mask_color,
                    mask_type=mask_type or "MultiColor"
                )
            )
        self.masks = masks


class BoundingBoxAnnotation(Annotation):
    def __init__(self, bounding_boxes: List[BoundingBox], **kwargs):
        super().__init__(annotation_type=AnnotationTypes.TWO_D_BOUNDING_BOX, **kwargs)
        self.bounding_boxes = bounding_boxes
        self.finalize()

    def set_annotation_entities(self):
        self.annotation_entities = []

        if not isinstance(self.bounding_boxes, list):
            raise InvalidAnnotationError(
                "bounding box annotation is expected to be list"
            )

        for bounding_box in self.bounding_boxes:
            if bounding_box.attributes:
                assert isinstance(bounding_box.attributes, dict), \
                    "attributes should be of type dict {} provided".format(type(bounding_box.attributes))
            if bounding_box.attributes:
                if bounding_box.attributes.get("track_id"):
                    assert isinstance(bounding_box.attributes["track_id"], int), \
                        "track_id should be of type int {} provided".format(type(bounding_box.attributes["track_id"]))
            try:
                category_id = bounding_box.category_id
                left = float(bounding_box.x_min)
                top = float(bounding_box.y_max)
                bottom = float(bounding_box.y_min)
                right = float(bounding_box.x_max)
            except KeyError as e:
                raise InvalidAnnotationError(str(e))

            coordinates = [
                {"x": left, "y": bottom},
                {"x": right, "y": bottom},
                {"x": right, "y": top},
                {"x": left, "y": top},
            ]
            self.annotation_entities.append(
                AnnotationEntity(
                    label=bounding_box.label,
                    category_id=category_id,
                    annotation_type=AnnotationTypes.TWO_D_BOUNDING_BOX,
                    coordinates=coordinates,
                    confidence=bounding_box.confidence,
                    attributes=bounding_box.attributes
                )
            )


class PolygonAnnotation(Annotation):
    def __init__(self, polygons: List[Polygon], **kwargs):
        super().__init__(annotation_type=AnnotationTypes.POLYGON, **kwargs)
        self.polygons = polygons
        self.finalize()

    def set_annotation_entities(self):
        self.annotation_entities = []

        if not isinstance(self.polygons, list):
            raise InvalidAnnotationError(
                "polygon annotation is expected to be list"
            )

        for polygon in self.polygons:
            try:
                category_id = polygon.category_id
                coords_ = polygon.coordinates
            except KeyError as e:
                raise InvalidAnnotationError(str(e))

            coordinates = []
            for c in coords_:
                coordinates.append({"x": c[0], "y": c[1]})
            self.annotation_entities.append(
                AnnotationEntity(
                    label=polygon.label,
                    category_id=category_id,
                    annotation_type=AnnotationTypes.POLYGON,
                    coordinates=coordinates,
                    confidence=polygon.confidence,
                    attributes=polygon.attributes
                )
            )


class PointAnnotation(Annotation):
    def __init__(self, points: List[Point], **kwargs):
        super().__init__(annotation_type=AnnotationTypes.POINT, **kwargs)
        self.points = points
        self.finalize()

    def set_annotation_entities(self):
        self.annotation_entities = []

        if not isinstance(self.points, list):
            raise InvalidAnnotationError(
                "point annotation is expected to be list"
            )

        for point in self.points:
            try:
                category_id = point.category_id
                left = float(point.left)
                top = float(point.top)

            except KeyError as e:
                raise InvalidAnnotationError(str(e))

            coordinates = [
                {"x": left, "y": top}
            ]

            self.annotation_entities.append(
                AnnotationEntity(
                    label=point.label,
                    category_id=category_id,
                    annotation_type=AnnotationTypes.POINT,
                    coordinates=coordinates,
                    confidence=point.confidence,
                    attributes=point.attributes
                )
            )


class LineAnnotation(Annotation):
    def __init__(self, lines: List[Line], **kwargs):
        super().__init__(annotation_type=AnnotationTypes.LINE, **kwargs)
        self.lines = lines
        self.finalize()

    def set_annotation_entities(self):
        self.annotation_entities = []

        if not isinstance(self.lines, list):
            raise InvalidAnnotationError(
                "line annotation is expected to be list"
            )

        for line in self.lines:
            try:
                category_id = line.category_id
                coords_ = line.coordinates
            except KeyError as e:
                raise InvalidAnnotationError(str(e))

            coordinates = []
            for c in coords_:
                coordinates.append({"x": c[0], "y": c[1]})

            self.annotation_entities.append(
                AnnotationEntity(
                    label=line.label,
                    category_id=category_id,
                    annotation_type=AnnotationTypes.LINE,
                    coordinates=coordinates,
                    confidence=line.confidence,
                    attributes=line.attributes
                )
            )


class ClassificationAnnotation(Annotation):
    def __init__(self, labels: List[Label], **kwargs):
        super().__init__(annotation_type=AnnotationTypes.CLASSIFICATION, **kwargs)
        self.labels = labels
        self.finalize()

    def set_annotation_entities(self):
        self.annotation_entities = []
        for label in self.labels:
            self.annotation_entities.append(
                AnnotationEntity(
                    label=label.label,
                    category_id=label.category_id,
                    class_name=label.class_name,
                    annotation_type=AnnotationTypes.CLASSIFICATION,
                    confidence=label.confidence
                )
            )


class PersonPose(Pose):
    default_keypoint_names = ['nose', 'left_eye', 'right_eye', 'left_ear', 'right_ear',
                              'left_shoulder', 'right_shoulder', 'left_elbow',
                              'right_elbow', 'left_wrist', 'right_wrist', 'left_hip',
                              'right_hip', 'left_knee', 'right_knee', 'left_ankle',
                              'right_ankle']

    def __new__(cls, keypoints: List[Point],
                keypoint_names: Optional[List[str]] = default_keypoint_names,
                attributes: Optional[dict] = None):

        label = "person"
        self = super(PersonPose, cls).__new__(cls, label,
                                              keypoints,
                                              keypoint_names,
                                              attributes)
        return self


class PoseAnnotation(Annotation):
    def __init__(self, poses: List[Pose], **kwargs):
        super().__init__(annotation_type=AnnotationTypes.TWO_D_POSE, **kwargs)
        self.poses = poses
        self.finalize()

    def set_annotation_entities(self):
        self.annotation_entities = []

        if not isinstance(self.poses, list):
            raise InvalidAnnotationError(
                "pose annotation is expected to be a list"
            )

        for pose in self.poses:
            assert len(pose.keypoint_names) == len(pose.keypoints)

            composition = []
            for i, keypoint_name in enumerate(pose.keypoint_names):
                try:
                    composition.append({
                        keypoint_name: pose.keypoints[i]
                    })
                except KeyError as e:
                    raise InvalidAnnotationError(str(e))

            self.annotation_entities.append(
                AnnotationEntity(
                    label=pose.label,
                    composition=composition,
                    annotation_type=AnnotationTypes.TWO_D_POSE,
                    attributes=pose.attributes
                )
            )




#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.
#

import io
import json
import os
import re
import shutil
import tempfile
import time
from copy import deepcopy
from datetime import datetime
from os.path import exists
from typing import Dict, List, Optional, Union, Any, Iterator
from pathlib import Path
from abc import abstractmethod

import requests
from diskcache import Cache

from ..constants import (AUXILIARY_KEY, TIMESTAMP_FIELD, AssetsConstants,
                         JobConstants, AggregationsByField, MAX_THREADS)
from ..custom_exceptions import (IdentityError, AssetError,
                                 InvalidAuthorization, JobError,
                                 MetadataNotFoundError)
from ..models.object_access import ObjectAccess
from ..tools import misc
from ..tools.connections import KVStoreTemplate
from ..tools.logger import get_logger
from ..tools.misc import parse_file_path, retry, tqdm
from ..tools.naming import standardize_name
from ..tools.time_utils import jsonify_metadata
from ..tools.misc import run_threaded, join

DEFAULT_SEARCH_SIZE = 50
DEFAULT_PAGINATION = 5000
PAUSE_TIME = 0.05

logger = get_logger(__name__)

VERSION_PATTERN = re.compile(r'/v\d+')
temp_dir = Path(tempfile.TemporaryDirectory().name).parent
DEFAULT_CACHE_DIR = os.getenv('CACHE_DIR', temp_dir)
DEFAULT_CACHE_SIZE = os.getenv("CACHE_MAX_SIZE", 50)
DEFAULT_CACHE_SUB_DIR = "default"
DEFAULT_ORGANIZATION_SUB_DIR = 'default'


class AssetManagerClientTemplate(object):
    """File cache for assets fetched via the AssetManagerClient.

    Using the least-frequently-used eviction policy, the cache will fill
    until the size_limit_gb is reached and will begin eviction
    """

    def __init__(
            self,
            asset_type: str,
            metadata_only: bool,
            cache_size_gb: int = None,
            disabled: bool = False,
            identity: Optional[dict] = None
    ):

        if not cache_size_gb:
            cache_size_gb = DEFAULT_CACHE_SIZE

        self.asset_type = asset_type
        self.metadata_only = metadata_only
        self.cache_disabled = disabled
        self.cache_size_gb = int(cache_size_gb)

        # Initialized in the _set_cache method
        self.cache_location = None
        self.cache = None
        self._set_cache(asset_type=asset_type, identity=identity)

    @abstractmethod
    def get_metadata(self, id: str, retry: bool = False) -> dict:
        pass

    @abstractmethod
    def get_url(self, id: str) -> str:
        pass

    @abstractmethod
    def get_bytes(self, id: str) -> bytes:
        pass

    def _get_cache_location(
            self,
            asset_type: Optional[str] = None,
            identity: Optional[dict] = None):
        # When not provided, defaults to <DEFAULT_CACHE_DIR,asset_type>.
        # If asset_type is None (for example, when initializing SceneEngineClient),
        # then defaults to a default sub directory.
        if asset_type:
            cache_location = os.path.join(DEFAULT_CACHE_DIR, asset_type)
        else:
            cache_location = os.path.join(
                DEFAULT_CACHE_DIR, DEFAULT_CACHE_SUB_DIR)

        if identity:
            organization_id = identity.get('organization_id', '')
        else:
            organization_id = None

        if organization_id:
            cache_location = os.path.join(cache_location, organization_id)
        else:
            cache_location = os.path.join(cache_location, DEFAULT_ORGANIZATION_SUB_DIR)

        self.cache_location = cache_location
        return cache_location

    def _set_cache(
            self,
            asset_type: Optional[str] = None,
            identity: Optional[dict] = None) -> None:
        """Set the cache location.

        Args:
            asset_type (str): Asset type
            identity (dict): User identity dict
        """
        try:
            self.cache_location = self._get_cache_location(
                asset_type=asset_type, identity=identity)
            os.makedirs(self.cache_location, exist_ok=True)
        except Exception as e:
            logger.info(
                " cache_location ::: could not create {} because of {}, instead {} is created".format(
                    self.cache_location, e, temp_dir))
            self.cache_location = temp_dir
            os.makedirs(self.cache_location, exist_ok=True)

    def get_file_from_cache(self, key):
        if self.cache_disabled:
            return
        # Read a file as a File object
        with Cache(
                directory=self.cache_location,
                eviction_policy='least-frequently-used',
                size_limit=self.cache_size_gb * (2 ** 30)
        ) as ref:
            return ref.get(key, read=True)

    def put_file_into_cache(self, key, value):
        if self.cache_disabled:
            return
        # Put a file as a File object
        with Cache(
                directory=self.cache_location,
                eviction_policy='least-frequently-used',
                size_limit=self.cache_size_gb * (2 ** 30)
        ) as ref:
            return ref.set(key, value, read=True)

    def get_bytes_from_cache(self, key):
        if self.cache_disabled:
            return
        # Read a file as a bytes array
        with Cache(
                directory=self.cache_location,
                eviction_policy='least-frequently-used',
                size_limit=self.cache_size_gb * (2 ** 30)
        ) as ref:
            return ref.get(key, read=False)

    def put_bytes_into_cache(self, key, value):
        if self.cache_disabled:
            return
        with Cache(
                directory=self.cache_location,
                eviction_policy='least-frequently-used',
                size_limit=self.cache_size_gb * (2 ** 30)
        ) as ref:
            return ref.set(key, value, read=False)


class Asset(object):
    def __init__(self,
                 id: str,
                 asset_manager_client: AssetManagerClientTemplate
                 ):
        self.id = id
        self.asset_manager_client = asset_manager_client
        self.asset_type = asset_manager_client.asset_type
        self.metadata_only = asset_manager_client.metadata_only

    def __check_metadata_only(self):
        if self.metadata_only:
            raise AssetError("{} {} is metadata only".format(self.asset_type, self.id))

    @property
    def metadata(self) -> Dict:
        return self.asset_manager_client.get_metadata(id=self.id)

    @property
    def url(self) -> str:
        self.__check_metadata_only()
        return self.asset_manager_client.get_url(id=self.id)

    @property
    def bytes(self) -> bytes:
        self.__check_metadata_only()
        return self.asset_manager_client.get_bytes(id=self.id)

    def download_as(self, filepath):
        with open(filepath, 'wb') as file:
            file.write(self.bytes)


class AssetManagerClient(AssetManagerClientTemplate):
    def __init__(
            self,
            asset_type: str = AssetsConstants.SESSIONS_ASSET_ID,  # default asset is session
            asset_manager_url: Optional[str] = None,
            metadata_only: bool = False,
            auth_token: Optional[str] = None,
            cache_disabled: bool = False,
            is_non_standard_asset: bool = False,
            kv_store: Optional[KVStoreTemplate] = None,
            num_threads: int = MAX_THREADS,
    ):
        """Asset manager client for fetching/putting assets and metadata.

        :param asset_type: (str) The type of asset
        :param asset_manager_url: (str) The URL of the asset manager
        :param metadata_only: (bool) Whether to only return metadata
        :param auth_token: (str) Authentication token of the user -- must be provided
        :param cache_disabled: (bool): Disable caching; defaults to False
        :param is_non_standard_asset (bool): Whether the asset is a non-standard asset.
               If True, ignore asset-type checks.
        :param kv_store: key-value store to cache bytes, default to None.
        """

        if is_non_standard_asset is False and asset_type not in AssetsConstants.VALID_ASSETS:
            raise ValueError('Invalid asset_type `{}`'.format(asset_type))

        # Override enabled cache with env variable
        if misc.as_bool(os.environ.get("DISABLE_CACHING", False)):
            cache_disabled = True

        self.is_non_standard_asset = is_non_standard_asset

        if asset_manager_url is None:
            raise AssetError('asset_manager_url must be provided')
        self.asset_manager_url = asset_manager_url
        self.jobs_namespace_url = join(asset_manager_url, "jobs")
        self.assets_namespace_url = join(asset_manager_url, "assets")

        self.auth_token = auth_token
        self.num_threads = num_threads
        if self.auth_token:
            self.identity = self._get_identity()
        else:
            self.identity = None

        if kv_store:
            self.kv_store = kv_store
        else:
            self.kv_store = None

        super().__init__(
            asset_type=asset_type,
            metadata_only=metadata_only,
            cache_size_gb=DEFAULT_CACHE_SIZE,
            disabled=cache_disabled,
            identity=self.identity
        )

        self._set_asset_state(
            asset_type=asset_type,
            metadata_only=metadata_only)

    def _get_root_url(self):
        # Strip version from asset manager URL
        version_match = re.search(VERSION_PATTERN, self.asset_manager_url)
        if version_match:
            version = version_match.group()
            root_url = self.asset_manager_url.split(version)[0]
        else:
            root_url = self.asset_manager_url
        return root_url

    def _get_identity(
            self,
            wait_for_connection: bool = True,
            max_wait_time: int = 60):
        env_user_identity = os.getenv('SCENEBOX_USER_IDENTITY')
        if env_user_identity:
            # TODO: This is a temporary fix and should be improved
            return json.loads(env_user_identity)
        else:
            if self.auth_token is None:
                raise IdentityError(
                    'Cannot get identity when no auth has been provided')
            params = {}
            if self.auth_token:
                params['token'] = self.auth_token

            root_url = self._get_root_url()

            wait_time = 0
            sleep_time = 5
            while True:
                try:
                    resp = requests.get(
                        join(
                            root_url,
                            'auth'),
                        params=params)
                    break
                except requests.exceptions.ConnectionError as e:
                    if not wait_for_connection or wait_time > max_wait_time:
                        raise e
                    else:
                        logger.warning(
                            "could not connect to {}, retrying in {} seconds".format(
                                root_url, sleep_time))
                        time.sleep(sleep_time)
                        wait_time = wait_time + sleep_time
            if not resp.ok:
                raise IdentityError(
                    "Could not obtain the user identity {}", resp.content)
            return resp.json()

    def _set_asset_state(
            self,
            asset_type: str,
            metadata_only: bool
    ):
        if asset_type not in AssetsConstants.VALID_ASSETS:
            raise AssetError(
                "The asset type {} is invalid. Valid types are: {}".format(
                    asset_type, AssetsConstants.VALID_ASSETS))

        if not isinstance(metadata_only, bool):
            raise ValueError("metadata_only must be a boolean")

        self.asset_type = asset_type

        if asset_type in AssetsConstants.METADATA_ONLY_ASSETS:
            self.metadata_only = True
        else:
            self.metadata_only = metadata_only

        if metadata_only is False:
            self._set_cache(asset_type=asset_type, identity=self.identity)
        return self

    def add_asset_manager_params(
            self, params_dict: Optional[dict] = None) -> dict:

        if self.auth_token is None:
            raise InvalidAuthorization(
                "No authorization is provided in AssetManagerClient")

        params_dict = params_dict or {}
        params_dict['asset_type'] = self.asset_type
        params_dict['metadata_only'] = self.metadata_only
        if self.auth_token:
            params_dict['token'] = self.auth_token
        return params_dict

    def with_auth(self,
                  auth_token: Optional[str] = None,
                  ):
        if auth_token:
            self.auth_token = auth_token

        self.identity = self._get_identity()
        self._set_cache(asset_type=self.asset_type, identity=self.identity)
        return self

    def register_field(self, field_dict: dict):
        url = self.get_assets_url('register_field/')
        params = self.add_asset_manager_params({})
        resp = requests.post(
            url,
            json=field_dict,
            params=params)
        if not resp.ok:
            raise AssetError(
                'Could not register the field ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        return resp.json()

    def delete(
            self,
            id: str,
            wait_for_deletion: bool = True) -> dict:
        """Delete an asset.

        Parameters
        ----------
        id:
            The ID of the asset to delete.
        wait_for_deletion:
            If True, polls until the specified asset no longer exists.
            Otherwise, returns immediately (even if the asset is still in the process
            of being deleted).

        """

        id = self._check_identifier(id)
        url = self.get_assets_url(id)
        params = self.add_asset_manager_params({})
        resp = requests.delete(url, params=params)

        if not resp.ok:
            raise AssetError(
                'Could not delete the file ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))

        if wait_for_deletion:
            self._wait_for_non_existence(id)

        return resp.json()

    def put_directory(self,
                      directory_path: str,
                      metadata: dict,
                      temp_dir=temp_dir) -> str:
        """Convert a directory to a zipfile and upload."""
        if not os.path.exists(directory_path):
            raise IOError('Path does not exist')

        with tempfile.NamedTemporaryFile(dir=temp_dir) as f:
            zip_ext = 'zip'
            temp_zipfile_path = f.name
            shutil.make_archive(temp_zipfile_path, zip_ext, directory_path)

            return self.put_asset(
                file_path=temp_zipfile_path,
                metadata=metadata,
                id="{}.{}".format(temp_zipfile_path, zip_ext),
                wait_for_completion=True)

    def put_file(
            self,
            file_object: Union[io.BytesIO, bytes, str],
            id: Optional[str] = None,
            owner_organization_id: Optional[str] = None,
            content_type: Optional[str] = None,
            content_size: Optional[int] = None,
            add_to_redis_cache: bool = False,
            wait_for_completion=False) -> ObjectAccess:
        """ Register and upload a file with the asset manager
        Args:
            file (BytesIO or bytes): bytes object
            id (str): the uid of the file. If not provided, will be set automatically
            owner_organization_id (str): file owner organization id
            wait_for_completion (bool): Whether to wait for the upload to finish before returning
        :return: object_access for the uploaded file
        """

        if id is not None:
            # delete asset if id exists
            params_dict = {"owner_organization_id": owner_organization_id} if owner_organization_id else {}
            params_for_delete = self.add_asset_manager_params(params_dict=params_dict)

            resp = requests.delete(
                self.get_assets_url('delete_if_exists_in_storage/{}'.format(id)),
                params=params_for_delete)
            if not resp.ok:
                raise AssetError(f"could not delete an existing asset {id} -- {resp.reason}, {resp.content}")
        else:
            id = misc.get_guid()


        large_size_bytes = 50 * 1024 * 1024  # 50 mb
        if content_size:
            file_size = content_size
        else:
            try:
                file_size = os.fstat(file_object.fileno()).st_size
            except:
                file_size = 0

        # single part upload
        if file_size < large_size_bytes:

            if isinstance(file_object, (io.BytesIO, io.IOBase)):
                file_object.seek(0)
                file = file_object.read()
            elif isinstance(file_object, str):
                file = file_object.encode("utf-8")
            else:
                file = file_object

            upload_params = self.get_temporary_upload_url(
                ids=[id],
                owner_organization_id=owner_organization_id,
                content_type=content_type
            )
            upload_url = upload_params[id]["url"]
            files = upload_params[id]["fields"]
            files["file"] = file

            resp = requests.post(upload_url, files=files)
            if not resp.ok:
                raise AssetError(
                    'Could not put the file ::: {} -- {} -- {}'.format(
                        self.asset_type, resp.reason, resp.content))
            if add_to_redis_cache:
                self.put_bytes_into_kv_store(id=id, data_bytes=file_object)
            object_access = ObjectAccess(uri=upload_params[id]["uri"])

        # multipart upload
        else:
            if not isinstance(file_object, (io.BytesIO, io.IOBase)):
                raise NotImplementedError("type should be Bytes or IOBase")
            params = self.add_asset_manager_params()
            body_complete_multipart = {"id": id}

            if owner_organization_id:
                body_complete_multipart["owner_organization_id"] = owner_organization_id

            resp = requests.post(
                self.get_assets_url('multipart_upload_id/'),
                json=body_complete_multipart,
                params=params)
            if not resp.ok:
                raise AssetError("could not get multipart upload id")
            upload_id = resp.json()["upload_id"]

            parts = []
            part_no = 1

            file_object.seek(0)
            pbar = tqdm(total=int(file_size/(1024 * 1024)), desc="uploading {} id:{} of size {:6.2f} MB".format(
                self.asset_type,
                id,
                file_size/(1024 * 1024)
            ))
            while True:
                file_data = file_object.read(large_size_bytes)
                if not file_data:
                    break

                body_complete_multipart = {"id": id, "part_no": part_no, "upload_id": upload_id}
                if owner_organization_id:
                    body_complete_multipart["owner_organization_id"] = owner_organization_id

                resp = requests.post(
                    self.get_assets_url('multipart_temporary_upload_url/'),
                    json=body_complete_multipart,
                    params=params)
                if not resp.ok:
                    raise AssetError("could not get multipart upload url")
                signed_url = resp.json()["url"]

                res = requests.put(signed_url, data=file_data)
                etag = res.headers['ETag']
                parts.append({'ETag': etag, 'PartNumber': part_no})
                part_no += 1
                pbar.update(int(large_size_bytes/(1024 * 1024)))
            pbar.close()

            body_complete_multipart = {"id": id, "parts": parts, "upload_id": upload_id}
            if owner_organization_id:
                body_complete_multipart["owner_organization_id"] = owner_organization_id

            resp = requests.post(
                self.get_assets_url('multipart_complete/'),
                json=body_complete_multipart,
                params=params)
            if not resp.ok:
                raise AssetError("could not get multipart upload url")

            file_object.close()
            uri = resp.json()["uri"]

            object_access = ObjectAccess(uri=uri)

        return object_access

    def get_url(self,
                id: str) -> str:
        """Get the public URL of an asset.

        Parameters
        ----------
        id:
            The ID of the asset to get the URL of.

        Returns
        -------
        str
            The URL of the specified asset.
        """
        id = self._check_identifier(id)
        url = self.get_assets_url(id)
        params = self.add_asset_manager_params({"return_url": True})
        resp = requests.get(url, params=params)
        if not resp.ok:
            raise AssetError(
                'Could not get the file ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        return resp.json()['url']

    @staticmethod
    def get_key_with_prefix(prefix: str, asset_id: str):
        return "{}_{}".format(prefix, asset_id)

    def put_bytes_into_kv_store(self, id: str, data_bytes: bytes):
        if self.kv_store:
            try:
                self.kv_store.put_bytes(key=self.get_key_with_prefix(prefix=self.asset_type, asset_id=id), value=data_bytes)
            except Exception as e:
                logger.warning("{}".format(e))

    def get_bytes_from_kv_store(self, id: str) -> Optional[bytes]:
        """Get the data bytes from kv_store if available.
        Parameters
        ----------
        id:
            The ID of the asset to get the bytes of.
        """
        if self.kv_store:
            return self.kv_store.get_bytes(key=self.get_key_with_prefix(prefix=self.asset_type, asset_id=id))
        else:
            return None

    def get_bytes(self,
                  id: str) -> bytes:
        """Get the data bytes of an asset.

        Parameters
        ----------
        id:
            The ID of the asset to get the bytes of.

        Returns
        -------
        bytes
            The data bytes of the specified asset.
        """
        id = self._check_identifier(id)

        # Check if the response object exists in kv_store
        data_bytes = self.get_bytes_from_kv_store(id)

        # Check if the response object exists locally in the cache
        if not data_bytes:
            data_bytes = self.get_bytes_from_cache(id)

        if not data_bytes:
            self._check_storage()
            url = self.get_assets_url(id)
            params = self.add_asset_manager_params({"return_url": False})
            resp = requests.get(url, params=params)
            if not resp.ok:
                raise AssetError(
                    'Could not get asset {} with id {}. Reason: {}: Content: {}'.format(
                        self.asset_type, id, resp.reason, resp.content))
            data_bytes = resp.content
            self.put_bytes_into_kv_store(id, data_bytes)
            self.put_bytes_into_cache(id, data_bytes)

        return data_bytes

    def get_bytes_in_batch(self,
                           ids: List[str]) -> Dict[str, bytes]:
        """Get the data bytes for a list of assets.

        Parameters
        ----------
        ids:
            The IDs of the assets to get the data bytes of.

        Returns
        -------
        dict[str, str]:
            The URLs of the specified assets.  Keys are the inputted ``ids``.  Values are the corresponding data bytes.

        """

        data_id_bytes_map = {}

        def threaded_get_bytes_and_append(id: str):
            data_id_bytes_map[id] = self.get_bytes(id)

        run_threaded(func=threaded_get_bytes_and_append,
                     iterable=ids,
                     num_threads=self.num_threads,
                     desc="get bytes")

        return data_id_bytes_map

    def get_url_in_batch(self,
                         ids: List[str]) -> Dict[str, str]:
        """Get the public URLs of a list of assets.

        Parameters
        ----------
        ids:
            The IDs of the assets to get the URLs of.

        Returns
        -------
        dict[str, str]:
            The URLs of the specified assets.  Keys are the inputted ``ids``.  Values are the corresponding URLs.

        """

        _ids = [self._check_identifier(_id) for _id in ids]

        assets_meta = self.search_meta_large(query={
            "filters": [
                {
                    "field": "id",
                    "values": _ids
                }
            ]
        })

        if len(assets_meta) != len(ids):
            existing_ids = [asset_meta.get("id") for asset_meta in assets_meta]
            logger.warning("{} assets do not exist with ids ::: {}".format(len(ids) - len(ids),
                                                                           set(ids).difference(set(existing_ids))))
        object_accesses = []
        ids_from_meta = []
        ids_from_meta_with_no_object_access = []

        for asset_meta in assets_meta:
            if asset_meta.get("object_access"):
                object_accesses.append(asset_meta["object_access"])
                ids_from_meta.append(asset_meta["id"])
            else:
                ids_from_meta_with_no_object_access.append(asset_meta["id"])

        if ids_from_meta_with_no_object_access:
            logger.warning("{} assets do not have object access with ids ::: {}".format(len(ids_from_meta_with_no_object_access),
                                                                           ids_from_meta_with_no_object_access))
        body = {
            "object_accesses": object_accesses,
            "ids": ids_from_meta
        }

        params = self.add_asset_manager_params()

        resp = requests.post(
            self.get_assets_url('url_from_object_access_in_batch/'),
            json=body,
            params=params
        )
        if not resp.ok:
            raise AssetError('Could not get urls for {}::: {} -- {} -- {}'.format(
                _ids, self.asset_type, resp.reason, resp.content))

        return resp.json()

    def get_file_and_write(self,
                           id: str,
                           write_filepath: Optional[str] = None) -> str:

        bytesio_data = io.BytesIO(self.get_bytes(id=id))
        if not write_filepath:
            write_filepath = tempfile.NamedTemporaryFile(delete=False).name
        with open(write_filepath, 'wb') as fout:
            fout.write(bytesio_data.read())

        return write_filepath

    def exists(self,
               id: str) -> bool:
        """Check if an asset exists.

        Parameters
        ----------
        id:
            The ID of the asset to check the existence of.

        Returns
        -------
        bool
            Returns True if the named asset exists.  Otherwise, returns False.
        """
        id = self._check_identifier(id)
        url = self.get_assets_url('exists/{}'.format(id))
        params = self.add_asset_manager_params()
        resp = requests.get(url, params=params)
        if not resp.ok:
            raise AssetError(
                'Could not find the identifier ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        return misc.as_bool(resp.json())

    def exists_multiple(self,
                        ids: List[str]) -> List[bool]:
        """Check if an asset exists.

        Parameters
        ----------
        ids:
            List of the IDs of assets to check the existence of.

        Returns
        -------
        bool
            Returns True if the named assets exist.  Otherwise, returns False.
        """

        ids = [self._check_identifier(id) for id in ids]
        metadata = self.get_metadata_in_batch(ids=ids)
        metadata_keys = metadata.keys()

        return [id in metadata_keys and metadata[id] is not None for id in ids]


    def copy(self,
             id: str,
             new_id: str):
        """Copy an asset.

        Parameters
        ----------
        id:
            The ID of the asset to copy.
        new_id:
            The ID to give to the created asset copy.
        """
        id = self._check_identifier(id)
        url = self.get_assets_url('copy/{}'.format(id))
        params = self.add_asset_manager_params({"new_id": new_id})
        resp = requests.put(url, params=params)
        if not resp.ok:
            raise AssetError(
                'Could not copy the identifier {} to {} ::: {} -- {} -- {}'.format(
                    id, new_id, self.asset_type, resp.reason, resp.content))
        return resp.json()

    def count(self, search: Optional[dict] = None) -> int:
        """Count the number of assets satisfying a query.

        Parameters
        ----------
        search:
            Query to locate the data subset of interest.  Filters through existing
            assets according to the dictionary passed.

        Returns
        -------
        int
            The number of assets that fulfil ``search``.
        """
        url = self.get_assets_url('count/')
        params = self.add_asset_manager_params()
        search = search or {}
        resp = requests.post(
            url,
            json=search,
            params=params)
        if not resp.ok:
            raise AssetError(
                'Could not count the files::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        return resp.json()

    def _aggregate_by_field(self, aggregation: str, field_name: str, search: Optional[dict] = None) -> float:
        '''
        Args:
            aggregation: aggregation function
            field_name: the field to do aggregation over it
            search: search criteria for the assets

        Returns: aggregation result

        '''
        if aggregation not in AggregationsByField.SUPPORTED:
            raise ValueError(
                "{} aggregation is not supported".format(aggregation))

        if not field_name:
            raise ValueError(
                "field_name must be provided, {} is given".format(field_name))
        url = self.get_assets_url('{}/'.format(aggregation))
        params = self.add_asset_manager_params()
        search = search or {}
        aggregation_request = {
            'search': search,
            'field_name': field_name
        }
        resp = requests.post(
            url,
            json=aggregation_request,
            params=params)
        if not resp.ok:
            raise AssetError(
                'Could not obtain {} aggregation for field {} ::: {} -- {} -- {}'.format(
                    aggregation, field_name, self.asset_type, resp.reason, resp.content))
        return resp.json()

    def sum(self, field_name: str, search: Optional[dict] = None) -> float:
        """Sum a specified field over assets that satisfy a query.

        Find the aggregate sum of ``field_name`` over all the assets specified by ``search``.

        Parameters
        ----------
        field_name:
            The field name in the filtered assets' metadata of which to sum over.
        search:
            Query to locate the data subset of interest.  Filters through existing
            assets according to the dictionary passed.

        Returns
        -------
        float
            The aggregate sum of the specified field name over the filtered assets.
       """
        return self._aggregate_by_field(aggregation=AggregationsByField.SUM, field_name=field_name, search=search)

    def average(self, field_name: str, search: Optional[dict] = None) -> float:
        """Average a specified field over assets that satisfy a query.

        Find the aggregate average of ``field_name`` over all the assets specified by ``search``.

        Parameters
        ----------
        field_name:
            The field name in the filtered assets' metadata of which to average over.
        search:
            Query to locate the data subset of interest.  Filters through existing
            assets according to the dictionary passed.

        Returns
        -------
        float
            The aggregate average of the specified field name over the filtered assets.
       """
        '''
        Args:
            field_name: the field to do average over it
            search: search criteria for the assets

        Returns: average

        '''
        return self._aggregate_by_field(aggregation=AggregationsByField.AVERAGE, field_name=field_name, search=search)

    def cardinality(self, field_name: str, string_field: bool=True, search: Optional[dict] = None) -> float:
        '''
        Args:
            field_name: the field to do average over it
            search: search criteria for the assets

        Returns: Count of distinct values

        '''
        if string_field:
            field_name += ".keyword"
        return self._aggregate_by_field(aggregation=AggregationsByField.CARDINALITY, field_name=field_name, search=search)

    @retry(Exception, logger=logger)
    def get_metadata(self, id: str, retry: bool = False) -> dict:
        """Fetch an asset's metadata.

        Parameters
        ----------
        id:
            The ID of the asset to receive the metadata of.
        retry:
            If True, retries the call for total_tries set in retry decorator

        Returns
        -------
        dict
            The metadata of the specified asset.
        """
        id = self._check_identifier(id)
        url = self.get_assets_url('meta/{}'.format(id))
        params = self.add_asset_manager_params()
        resp = requests.get(url, params=params)
        if not resp.ok:
            raise MetadataNotFoundError(
                'Could not get the metadata from the identifier ::: {} -- {} -- {} -- {}'.format(
                    id, self.asset_type, resp.reason, resp.content))
        return resp.json()

    @retry(Exception, logger=logger)
    def get_metadata_in_batch(self,
                              ids: List[str]) -> Dict[str, Optional[dict]]:
        """Fetch an asset's metadata.

        Parameters
        ----------
        ids:
            The IDs of the asset to receive the metadata of.

        Returns
        -------
        Dict: str->dict or None
            The metadata of the specified asset-set to None if not found.
        """

        _ids = [self._check_identifier(_id) for _id in ids]

        assets_meta = self.search_meta_large(query={
            "filters": [
                {
                    "field": "id",
                    "values": _ids
                }
            ]
        })

        assets_meta_dict = {}
        existing_ids = set()
        for asset_meta in assets_meta:
            existing_id = asset_meta.get("id")
            assets_meta_dict[existing_id] = asset_meta
            existing_ids.add(existing_id)

        if len(existing_ids) != len(_ids):
            extra_ids = set(_ids).difference(existing_ids)
            # logger.warning("{} assets do not exist with ids ::: {}".format(len(extra_ids), extra_ids))
            for extra_id in extra_ids:
                assets_meta_dict[extra_id] = None

        return assets_meta_dict

    @retry(Exception, logger=logger)
    def put_asset(self,
                  metadata: dict,
                  file_path: Optional[str] = None,
                  filename: Optional[str] = None,
                  file_object: Optional[Union[io.BytesIO, bytes, str]] = None,
                  url: Optional[str] = None,
                  id: Optional[str] = None,
                  uri: Optional[str] = None,
                  geo_field: Optional[str] = None,
                  shape_group_field: Optional[str] = None,
                  buffered_write: bool = False,
                  wait_for_completion: bool = True,
                  content_type: Optional[str] = None,
                  add_to_redis_cache: bool = False,
                  retry: bool = False) -> str:

        # metadata should be always provided
        # either of the following should be provided if the asset is not metadata only
        # url
        # file_path
        # file_object and filename

        metadata_copy = deepcopy(metadata)
        id = id or metadata_copy.get("id") or misc.get_guid()
        metadata_copy["id"] = id

        if not metadata_copy.get(TIMESTAMP_FIELD):
            metadata_copy[TIMESTAMP_FIELD] = datetime.utcnow()

        if self.asset_type in AssetsConstants.METADATA_ONLY_ASSETS:
            if any([filename, file_path, url, file_path, uri]):
                raise AssetError(
                    "asset {} does not accept file, filename, file_path, uri or, url".format(
                        self.asset_type))
        else:
            if not any([file_object, url, file_path, uri]) and self.asset_type != AssetsConstants.ANNOTATIONS_ASSET_ID:
                raise AssetError(
                    "file_object, url, file_path is required for asset {}".format(
                        self.asset_type))
            object_access = None
            if url:
                object_access = ObjectAccess(url=url,
                                             filename=filename)
                if uri:
                    raise AssetError(
                        "url {} cannot support uri".format(url))
            elif uri:
                object_access = ObjectAccess(uri=uri,
                                             filename=filename)
                if url:
                    raise AssetError(
                        "uri {} cannot support url".format(uri))
            elif file_path or file_object:
                if file_path:
                    if not exists(file_path):
                        raise AssetError(
                            "file_path {} does not exist".format(file_path))
                    filename, _, _, _ = parse_file_path(file_path)
                    file_object = open(file_path, 'rb')
                else:
                    filename = filename or id

                object_access = self.put_file(
                    file_object=file_object,
                    id=id,
                    content_type=content_type,
                    content_size=metadata_copy.get("num_bytes", 0),
                    add_to_redis_cache=add_to_redis_cache
                )

            if object_access:
                metadata_copy["object_access"] = object_access.to_dic()

        metadata_json = jsonify_metadata(metadata_copy)

        params = {
            "replace": True,
            "geo_field": geo_field,
            "shape_group_field": shape_group_field,
            "buffered_write": buffered_write,
            "replace_sets": False}
        params = self.add_asset_manager_params(params)
        resp = requests.put(
            url=self.get_assets_url("meta/{}".format(id)),
            json=metadata_json,
            params=params)
        if not resp.ok:
            raise AssetError(
                'Could not put the metadata for the identifier ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))

        if wait_for_completion:
            self._wait_for_existence(id)
        return id

    @retry(Exception, logger=logger, initial_wait=1.5, total_tries=10)
    def put_assets_batch(self,
                  metadata: List[dict],
                  file_paths: Optional[List[str]] = None,
                  filenames: Optional[List[str]] = None,
                  file_objects: Optional[List[Union[io.BytesIO, bytes, str]]] = None,
                  urls: Optional[List[str]] = None,
                  ids: Optional[List[str]] = None,
                  uris: Optional[List[str]] = None,
                  geo_field: Optional[str] = None,
                  shape_group_field: Optional[str] = None,
                  wait_for_completion: bool = True,
                  content_type: Optional[str] = None,
                  threading: bool = True,
                  disable_tqdm: bool = True,
                  add_to_redis_cache: bool = False,
                  retry: bool = False) -> List[str]:

        # metadata should be always provided
        # either of the following should be provided if the asset is not metadata only
        # url
        # file_path
        # file_object and filename

        key_json_meta_map = {}
        key_id_map = {}

        def threaded_file_writer(iterable):
            key, id, metadatum, filename, file_path, file_object, url, uri = iterable

            id = id or metadatum.get("id") or misc.get_guid()
            metadatum["id"] = id

            if not metadatum.get(TIMESTAMP_FIELD):
                metadatum[TIMESTAMP_FIELD] = datetime.utcnow()

            if self.asset_type in AssetsConstants.METADATA_ONLY_ASSETS:
                if any([filename, file_path, url, file_object, uri]):
                    raise AssetError(
                        "asset {} does not accept file, filename, file_path, uri or, url".format(
                            self.asset_type))
            else:
                if not any([file_object, url, file_path, uri]) and self.asset_type != AssetsConstants.ANNOTATIONS_ASSET_ID:
                    raise AssetError(
                        "file_object, url, file_path is required for asset {}".format(
                            self.asset_type))
                object_access = None
                if url:
                    object_access = ObjectAccess(url=url,
                                                 filename=filename)
                    if uri:
                        raise AssetError(
                            "url {} cannot support uri".format(url))
                elif uri:
                    object_access = ObjectAccess(uri=uri,
                                                 filename=filename)
                    if url:
                        raise AssetError(
                            "uri {} cannot support url".format(uri))
                elif file_path or file_object:
                    if file_path:
                        if not exists(file_path):
                            raise AssetError(
                                "file_path {} does not exist".format(file_path))
                        filename, _, _, _ = parse_file_path(file_path)
                        file_object = open(file_path, 'rb')
                    else:
                        filename = filename or id

                    object_access = self.put_file(
                        file_object=file_object,
                        id=id,
                        content_type=content_type,
                        content_size=metadatum.get("num_bytes", 0),
                        add_to_redis_cache=add_to_redis_cache)

                if object_access:
                    metadatum["object_access"] = object_access.to_dic()

            metadatum_json = jsonify_metadata(metadatum)
            key_json_meta_map[key] = metadatum_json
            key_id_map[key] = id

        # Formatting inputs to a list to run threaded
        if ids is None:
            ids = [None] * len(metadata)
        if filenames is None:
            filenames = [None] * len(metadata)
        if file_paths is None:
            file_paths = [None] * len(metadata)
        if file_objects is None:
            file_objects = [None] * len(metadata)
        if urls is None:
            urls = [None] * len(metadata)
        if uris is None:
            uris = [None] * len(metadata)

        keys = list(range(len(metadata)))
        # logger.debug("Writing {} files to disk.".format(len(metadata)))
        run_threaded(func=threaded_file_writer,
                     iterable=zip(keys, ids, metadata, filenames, file_paths,
                                  file_objects, urls, uris),
                     desc="adding annotations",
                     num_threads=self.num_threads,
                     disable_threading=not threading,
                     disable_tqdm=disable_tqdm)

        json_metadata_list = []
        ids_list = []
        for key in keys:
            json_metadata_list.append(key_json_meta_map[key])
            ids_list.append(key_id_map[key])

        # Rename to avoid confusion
        ids = ids_list
        body = {"ids": ids, "metadata": json_metadata_list}

        params = {
            "replace": True,
            "geo_field": geo_field,
            "shape_group_field": shape_group_field,
            "replace_sets": False}

        params = self.add_asset_manager_params(params)

        # logger.debug("Putting {} metadata in one batch.".format(len(json_metadata_list)))
        resp = requests.post(
            self.get_assets_url('meta_batch/'),
            json=body,
            params=params)

        if not resp.ok:
            raise AssetError(
                'Could not put the metadata for the identifier ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))

        if wait_for_completion:
            self._wait_for_existence(ids=ids)

        return ids

    @retry(Exception, logger=logger)
    def update_metadata(
            self,
            id: str,
            metadata: dict,
            buffered_write: bool = False,
            replace_sets: bool = False,
            geo_field: Optional[str] = None,
            shape_group_field: Optional[str] = None,
            retry: bool = False) -> dict:
        """Update an asset's metadata.

        Parameters
        ----------
        id:
            The ID of the asset to update the metadata of.
        metadata:
            The metadata to update the existing metadata with.
        buffered_write:
            If True, ingests the metadata in a buffered fashion.
        replace_sets:
            If True, replaces existing metadata with ``metadata``.  Otherwise, attempts to append ``metadata`` to the
            existing metadata.
        geo_field:
            Geolocation field
        shape_group_field:
            Shape group field (example: UMAP)
        retry:
            If True, retries the call for total_tries set in retry decorator
        """

        id = self._check_identifier(id)
        url = self.get_assets_url('meta/{}'.format(id))
        params = {"replace": False,
                  "replace_sets": replace_sets,
                  "buffered_write": buffered_write,
                  "geo_field": geo_field,
                  "shape_group_field": shape_group_field
                  }
        params = self.add_asset_manager_params(params)
        metadata_json = jsonify_metadata(metadata)

        resp = requests.put(
            url,
            json=metadata_json,
            params=params)
        if not resp.ok:
            raise AssetError(
                'Could not update the metadata for the identifier ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        return resp.json()

    @retry(Exception, logger=logger, initial_wait=1.5, total_tries=10)
    def update_metadata_batch(
            self,
            ids: List[str],
            metadata: List[dict],
            buffered_write: bool = False,
            replace_sets: bool = False,
            geo_field: Optional[str] = None,
            shape_group_field: Optional[str] = None) -> dict:
        """Update an asset's metadata.

        Parameters
        ----------
        ids:
            The IDs of the assets to update the metadata of.
        metadata:
            The metadata to update the existing metadata with.
        buffered_write:
            If True, ingests the metadata in a buffered fashion.
        replace_sets:
            If True, replaces existing metadata with ``metadata``.  Otherwise, attempts to append ``metadata`` to the
            existing metadata.
        geo_field:
            Geolocation field
        shape_group_field:
            Shape group field (example: UMAP)
        retry:
            If True, retries the call for total_tries set in retry decorator
        """

        ids = [self._check_identifier(id) for id in ids]
        url = self.get_assets_url('meta_batch')
        params = {
                  "replace": False,
                  "replace_sets": replace_sets,
                  "buffered_write": buffered_write,
                  "geo_field": geo_field,
                  "shape_group_field": shape_group_field
                  }
        params = self.add_asset_manager_params(params)
        metadata_json = [jsonify_metadata(meta) for meta in metadata]

        body = {"ids": ids, "metadata": metadata_json}

        resp = requests.post(
            url,
            json=body,
            params=params)

        if not resp.ok:
            raise AssetError(
                'Could not update the metadata for the identifier ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        return resp.json()

    def upsert_metadata(
            self,
            id: str,
            metadata: dict,
            geo_field: Optional[str] = None,
            shape_group_field: Optional[str] = None,
            buffered_write: bool = False) -> dict:
        """Upsert an asset's metadata.

        Inserts new fields of ``metadata`` if they do not already exist.  Otherwise, updates the existing field if it
        does exist.

        Parameters
        ----------
        id:
            The ID of the asset to upsert the metadata of.
        metadata:
            The metadata to upsert the existing metadata with.
        geo_field:
            Geolocation field
        shape_group_field:
            Shape group field (example: UMAP)
        buffered_write:
            If True, ingests metadata in a buffered fashion.
        """
        id = self._check_identifier(id)
        url = self.get_assets_url('meta/{}'.format(id))
        params = {"replace": True,
                  "replace_sets": True,
                  "buffered_write": buffered_write,
                  "geo_field": geo_field,
                  "shape_group_field": shape_group_field
                  }
        params = self.add_asset_manager_params(params)
        metadata_json = jsonify_metadata(metadata)

        resp = requests.put(
            url,
            json=metadata_json,
            params=params)
        if not resp.ok:
            raise AssetError(
                'Could not upsert the metadata for the identifier ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        return resp.json()

    def update_by_query(self,
                        update_field: str,
                        update_value: Union[str, int, float, bool],
                        limit: Optional[int] = None,
                        query: Optional[dict] = None) -> dict:
        '''

        Args:
            update_field: field to update
            update_value: update value
            limit: limit the number of docs in query result. <limit> number of docs will be randomly selected. If not
                    provided, all documents satisfying the query will be added to set.
            query: search query to limit the docs to update

        Returns:
           update task id. you can check the status of task using the jobs/query_task endpoint. once the task is complete
           you may need to clear the cache for associated assets to view the lasted changes
        '''
        url = self.get_assets_url('update_by_query')
        params = self.add_asset_manager_params()
        body = {"update_field": update_field, "update_value": update_value}
        if limit:
            body["limit"] = limit
        if query:
            body["search"] = query

        resp = requests.post(
            url,
            json=body,
            params=params)

        if not resp.ok:
            raise AssetError(
                'Could not update {}={} for query {} and limit ::: {} - {}'.format(
                    update_field,update_value,query,limit, resp.json(), resp.content))
        return resp.json()

    def assets_iterator(
            self,
            query: Optional[dict] = None,
            filters: Optional[dict] = None,
            sort_field: str = "id",
            sort_order: str = "asc") -> Iterator[Asset]:

        """Retrieve an iterator for Assets with a search query.

                Parameters
                ----------
                query:
                    Query to locate the data subset of interest.  Filters through existing
                    assets according to the dictionary passed.
                filters:
                    Filter names and values to append to ``query``.  Dict keys represent the existing filter names,
                    and dict values are the filter values.
                sort_field:
                    Filters for the specified name.
                sort_order:
                    Specifies the string sorting order. [asc. desc]

                Returns
                -------
                Iterator[Asset]
                    Iterator for assets

                """

        class AssetIterator(Iterator):
            # Constructor
            def __init__(self,
                         asset_manager_client: AssetManagerClient,
                         query_: dict,
                         ):
                self.query = query_
                self.asset_manager_client = asset_manager_client
                self.params = {
                    "size": DEFAULT_PAGINATION,
                    "offset": 0,
                    "scan": False,
                    "sort_field": sort_field,
                    "sort_order": sort_order
                }
                self.index = 0
                self.results = []
                self.len = 0
                self.search_after = None

            def __next__(self):
                if self.index >= self.len:
                    params = {"size": DEFAULT_PAGINATION,
                              "offset": 0,
                              "scan": False,
                              "sort_field": sort_field,
                              "sort_order": sort_order
                              }
                    if self.search_after is not None:
                        params["search_after"] = self.search_after
                    url = self.asset_manager_client.get_assets_url('')
                    params = self.asset_manager_client.add_asset_manager_params(params)
                    resp = requests.post(url, json=query, params=params)
                    if not resp.ok:
                        raise AssetError(
                            'Could not search the files ::: {} -- {} -- {}'.format(
                                self.asset_manager_client.asset_type, resp.reason, resp.content))
                    self.results = resp.json().get("results")
                    self.len = len(self.results)
                    if self.results:
                        self.search_after = self.results[-1]
                    else:
                        raise StopIteration
                    self.index = 0

                asset = Asset(id=self.results[self.index],
                              asset_manager_client=self.asset_manager_client)
                self.index += 1
                return asset

        query = self._construct_query(query, filters)

        return AssetIterator(
            asset_manager_client=self,
            query_=query
        )

    def search_assets(
            self,
            query: Optional[dict] = None,
            filters: Optional[dict] = None,
            size: int = DEFAULT_SEARCH_SIZE,
            offset: int = 0,
            sort_field: Optional[str] = None,
            sort_order: Optional[str] = None,
            scan: bool = False) -> List[str]:
        """Retrieve asset IDs with a search query.

        Returns the top ``size`` matching hits.  If a return of more than 10000 hits is desired, please use
        AssetManagerClient.search_assets_large().

        Parameters
        ----------
        query:
            Query to locate the data subset of interest.  Filters through existing
            assets according to the dictionary passed.
        filters:
            Filter names and values to append to ``query``.  Dict keys represent the existing filter names,
            and dict values are the filter values.
        size:
            Specifies the Elasticsearch search size.  The maximum number of hits to return.
            Has no effect if ``scan`` is False.
        offset:
            Specifies the Elasticsearch search offset.  The number of hits to skip.
            Has no effect if ``scan`` is False.
        sort_field:
            Filters for the specified name.
        sort_order:
            Specifies the Elasticsearch string sorting order.
        scan:
            If True, uses the Elasticsearch scan capability.
            Otherwise, uses the Elasticsearch search API.

        Returns
        -------
        List[str]
            A list of the IDs of the assets fulfilling the search query.

        """
        query = self._construct_query(query, filters)
        params = {'size': size, 'offset': offset, 'scan': scan}
        if sort_field:
            params['sort_field'] = sort_field
        if sort_order:
            params['sort_order'] = sort_order
        url = self.get_assets_url('')
        params = self.add_asset_manager_params(params)
        resp = requests.post(url, json=query, params=params)
        if not resp.ok:
            raise AssetError(
                'Could not search the files ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        return resp.json()['results']

    def search_assets_large(
            self,
            query: Optional[dict] = None,
            filters: Optional[dict] = None
    ) -> List[str]:
        """Retrieve all asset IDs matching a search query.

        Return all hits matching a search query.  If a return of less than 10000 hits is desired, please use
        AssetManagerClient.search_assets() or SceneEngineClient.search_assets().

        Parameters
        ----------
        query:
            Query to locate the data subset of interest.  Filters through existing
            assets according to the dictionary passed.
        filters:
            Filter names and values to append to ``query``.  Dict keys represent the existing filter names,
            and dict values are the filter values.

        Returns
        -------
        List[str]
            A list of the IDs of the assets fulfilling the search query.
        """
        query = self._construct_query(query, filters)
        params = {"size": DEFAULT_PAGINATION,
                  "offset": 0,
                  "scan": False,
                  "sort_field": "id",
                  "sort_order": "asc"
                  }
        url = self.get_assets_url('')
        params = self.add_asset_manager_params(params)
        resp = requests.post(url, json=query, params=params)
        if not resp.ok:
            raise AssetError(
                'Could not search the files ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        results = resp.json().get('results')
        if results:
            search_after = resp.json()['results'][-1]
        else:
            return results
        while search_after is not None:
            params = {"size": DEFAULT_PAGINATION,
                      "offset": 0,
                      "scan": False,
                      "sort_field": "id",
                      "sort_order": "asc",
                      "search_after": search_after
                      }
            url = self.get_assets_url('')
            params = self.add_asset_manager_params(params)
            resp = requests.post(url, json=query, params=params)
            if not resp.ok:
                raise AssetError(
                    'Could not search the files ::: {} -- {} -- {}'.format(
                        self.asset_type, resp.reason, resp.content))
            if resp.json().get('results'):
                res = resp.json()['results']
                results += res
                search_after = res[-1]
            else:
                break
        return results

    def delete_with_query(self,
                          query: Optional[dict] = None,
                          filters: Optional[dict] = None,
                          wait_for_completion: bool = False) -> str:
        """Delete assets specified with a search query.

        Parameters
        ----------
        query:
            Query to locate the data subset of interest.  Filters through existing
            assets according to the dictionary passed.
        filters:
            Filter names and values to append to ``query``.  Dict keys represent the existing filter names,
            and dict values are the filter values.
        wait_for_completion:
            If True, polls until job is complete.
            Otherwise, continues execution and does not raise an error if the job does not complete.

        Returns
        -------
        str
            The id of the Job that carries out the asset deletion.

        """
        query = self._construct_query(query, filters)
        params = self.add_asset_manager_params()
        url = self.get_assets_url('remove_with_query/')
        resp = requests.post(url, json=query, params=params)
        if not resp.ok:
            raise AssetError(
                'Could not remove the assets ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        job_id = resp.json()['job_id']
        if wait_for_completion:
            self._wait_for_job_completion(job_id)

        return job_id

    def delete_with_list(self,
                         assets_list: Optional[list] = None,
                         wait_for_completion: bool = False) -> str:
        """Delete assets specified with a list of asset IDs.

        Parameters
        ----------
        assets_list:
            Asset IDs of the assets to delete.
        wait_for_completion:
            If True, polls until job is complete.
            Otherwise, continues execution and does not raise an error if the job does not complete.

        Returns
        -------
        str
            The id of the Job that carries out the asset deletion.
        """
        assets_list = assets_list or []
        params = self.add_asset_manager_params()
        url = self.get_assets_url('remove_with_list/')
        resp = requests.post(
            url,
            json={
                "assets_list": assets_list},
            params=params)
        if not resp.ok:
            raise AssetError(
                'Could not remove the assets ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))

        job_id = resp.json()['job_id']
        if wait_for_completion:
            self._wait_for_job_completion(job_id)

        return job_id

    def search_meta(
            self,
            query=None,
            size=DEFAULT_SEARCH_SIZE,
            filters: Optional[dict] = None,
            offset=0,
            sort_field=None,
            sort_order=None,
            scan=False,
            compress=False) -> List[dict]:

        """Retrieve asset metadata with a search query.

        Returns the top ``size`` matching hits.  If a return of more than 10000 hits is desired, please use
        AssetManagerClient.search_meta_large().

        Parameters
        ----------
        query:
            Query to locate the data subset of interest.  Filters through existing
            assets according to the dictionary passed.
        filters:
            Filter names and values to append to ``query``.  Dict keys represent the existing filter names,
            and dict values are the filter values.
        size:
            Specifies the Elasticsearch search size.  The maximum number of hits to return.
            Has no effect if ``scan`` is False.
        offset:
            Specifies the Elasticsearch search offset.  The number of hits to skip.
            Has no effect if ``scan`` is False.
        sort_field:
            Filters for the specified name.
        sort_order:
            Specifies the Elasticsearch string sorting order.
        scan:
            If True, uses the Elasticsearch scan capability.
            Otherwise, uses the Elasticsearch search API.
        compress:
            Boolean. If set to True, a gzip compressed list of metadata is returned.
            Typically used in cases where the metadata returned is over 20MB.

        Returns
        -------
        List[dict]
            A list of the metadata dicts of each asset that fulfills the search query.
        """
        query = self._construct_query(query, filters)
        params = {'size': size, 'offset': offset, 'scan': scan}
        if sort_field:
            params['sort_field'] = sort_field
        if sort_order:
            params['sort_order'] = sort_order
        if compress:
            params["compress"] = compress
        url = self.get_assets_url('meta')
        params = self.add_asset_manager_params(params)
        resp = requests.post(url, json=query, params=params)
        if not resp.ok:
            raise AssetError(
                'Could not search the metadata ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        return resp.json()

    def search_meta_large(
            self,
            query: Optional[dict] = None,
            filters: Optional[dict] = None,
            compress: bool = False) -> List[dict]:
        """Retrieve all asset metadata matching a search query.

        Return all hits matching a search query.  If a return of less than 10000 hits is desired, please use
        AssetManagerClient.search_meta() or SceneEngineClient.search_meta().

        Parameters
        ----------
        query:
            Query to locate the data subset of interest.  Filters through existing
            assets according to the dictionary passed.
        filters:
            Filter names and values to append to ``query``.  Dict keys represent the existing filter names,
            and dict values are the filter values.
        compress:
            Boolean. If set to True, a gzip compressed list of metadata is returned.
            Typically used in cases where the metadata returned is over 20MB.

        Returns
        -------
        List[dict]
            A list of the metadata dicts of each asset that fulfills the search query.

        """
        query = self._construct_query(query, filters)
        params = {"size": DEFAULT_PAGINATION,
                  "offset": 0,
                  "scan": False,
                  "sort_field": "id",
                  "sort_order": "asc",
                  "compress": compress
                  }
        url = self.get_assets_url('meta')
        params = self.add_asset_manager_params(params)
        resp = requests.post(url, json=query, params=params)
        if not resp.ok:
            raise AssetError(
                'Could not search the files ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        results = resp.json()
        if results:
            search_after = resp.json()[-1]['id']
        else:
            return results
        while search_after:
            params = {"size": DEFAULT_PAGINATION,
                      "offset": 0,
                      "scan": False,
                      "sort_field": "id",
                      "sort_order": "asc",
                      "search_after": search_after
                      }
            url = self.get_assets_url('meta')
            params = self.add_asset_manager_params(params)
            resp = requests.post(url, json=query, params=params)
            if not resp.ok:
                raise AssetError(
                    'Could not search the files ::: {} -- {} -- {}'.format(
                        self.asset_type, resp.reason, resp.content))
            if resp.json():
                res = resp.json()
                results += res
                search_after = res[-1]['id']
            else:
                break
        return results

    def summary_meta(self, summary_request: dict) -> dict:
        """Get a metadata summary.

        Parameters
        ----------
        summary_request:
            Dict of summary settings of the following form:
            {
                "search" (dict): <Query to locate the data subset of interest>,
                "dimensions" (Optional[List[str]): <Dimensions of summary (term aggregations)>,
                "custom_buckets_numeric" Optional[List[dict]]: <Buckets for range aggregation>
                "custom_buckets_time" Optional[List[dict]]: <Buckets for date_range aggregation>
                "aggregate_field" Optional[str]: <Field for metric aggregations>
                "aggregation_type" Optional[str]: <Type of metric aggregation (e.g. avg, sum)>
                "max_size_for_agg" Optional[str]: < Maximum number of buckets to return (default is 100)>
            }

        Returns
        -------
        dict
            Metadata summary according to ``summary_request``.

        """
        url = self.get_assets_url('meta/summary/')
        params = self.add_asset_manager_params()
        resp = requests.post(
            url,
            json=summary_request,
            params=params)
        if not resp.ok:
            raise AssetError(
                'Could not retrieve the metadata summary ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        return resp.json()

    def summary_2d(self, summary_2d_request):
        url = self.get_assets_url('summary_2d/')
        params = self.add_asset_manager_params()
        resp = requests.post(
            url,
            json=summary_2d_request,
            params=params)
        if not resp.ok:
            raise AssetError(
                'Could not get the confusion matrix for ::: {} -- {} -- {}'.format(
                    summary_2d_request, resp.reason, resp.content))
        return resp.json()

    def summary_nd(self, summary_nd_request: dict):
        url = self.get_assets_url('summary_nd/')
        params = self.add_asset_manager_params()
        resp = requests.post(
            url,
            json=summary_nd_request,
            params=params)
        if not resp.ok:
            raise AssetError(
                'Could not get the confusion matrix for ::: {} -- {} -- {}'.format(
                    summary_nd_request, resp.reason, resp.content))
        return resp.json()

    def api_status(self):
        url = self.get_assets_url('status/')
        params = self.add_asset_manager_params()
        resp = requests.get(url, params=params)
        if not resp.ok:
            raise AssetError(
                'Could not check the status of the API ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))

    def get_assets_url(self, route=""):
        url = join(self.assets_namespace_url, route)
        return url

    def get_statistics(self, field_dicts: List[dict], max_size_for_agg: Optional[int] = None):
        statistics_endpoint = self.get_assets_url("field_statistics/")
        params = self.add_asset_manager_params()
        payload = {'field_dicts': field_dicts}
        if max_size_for_agg is not None:
            payload["max_size_for_agg"] = max_size_for_agg
        resp = requests.post(
            statistics_endpoint,
            params=params,
            json=payload)
        if not resp.ok:
            raise AssetError(
                'Could not get field statistics ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        return resp.json()

    def _get_job_status(self, job_id):
        """Get the status of the async job.

        Assumes that the job_id is the same as the celery task_id

        :param job_id: Celery task ID
        :return:
        """
        status_endpoint = join(
            self.jobs_namespace_url, "status", job_id)
        params = {}
        if self.auth_token:
            params['token'] = self.auth_token
        resp = requests.get(status_endpoint, params=params)
        status = resp.json().get('status')
        return status

    def _wait_for_job_completion(self, job_id):
        while True:
            status = self._get_job_status(job_id)
            if status == JobConstants.STATUS_FINISHED:
                break
            elif status in {JobConstants.STATUS_ABORTED}:
                job_metadata = self.with_asset_state(
                    'jobs', True).get_metadata(job_id)
                job_notes = job_metadata['notes']
                raise JobError(
                    'Job {} encountered error with status {} with notes::: {}'.format(
                        job_id, status, job_notes))
            else:
                time.sleep(1)

    def _wait_for_existence(self, id = None,
                            ids: List[str] = None,
                            increments_sec: int = 5,
                            max_wait_sec: int = 120):
        elapsed_time = 0
        if ids is not None:
            # Wait on multiple assets
            exists_status = [False] * len(ids)
            while sum(exists_status) < len(ids):
                if elapsed_time >= max_wait_sec:
                    raise TimeoutError("Timeout waiting for {} ids to exist".format(len(ids)))

                exists_status = self.exists_multiple(ids)
                elapsed_time += increments_sec
                time.sleep(increments_sec)
        else:
            # Wait on single asset
            id = self._check_identifier(id)
            while not self.exists(id):
                time.sleep(PAUSE_TIME)

    def _wait_for_non_existence(self, id):
        id = self._check_identifier(id)
        while self.exists(id):
            time.sleep(PAUSE_TIME)

    def _check_storage(self):
        if self.metadata_only:
            raise AssetError('Asset client is metadata only')

    def update_asset_state(self, asset_type: str, metadata_only: bool = False):

        if self.is_non_standard_asset is False and asset_type not in AssetsConstants.VALID_ASSETS:
            raise AssetError(
                "The asset type {} is invalid. Valid types are: {}".format(
                    asset_type, AssetsConstants.VALID_ASSETS))

        if not isinstance(metadata_only, bool):
            raise ValueError("metadata_only must be a boolean")

        self.asset_type = asset_type
        if asset_type in AssetsConstants.METADATA_ONLY_ASSETS:
            self.metadata_only = True
        else:
            self.metadata_only = metadata_only

    def with_asset_state(self, asset_type: str, metadata_only: bool = False):
        """Set the asset state, for use in chaining.

        Eg. client.with_asset_state("images", True).search_files({})
        """

        self.update_asset_state(
            asset_type=asset_type,
            metadata_only=metadata_only)
        return self

    def _construct_query(self,
                         query: Optional[Dict],
                         filters: Optional[Dict]):
        if query is None:
            query = {}
        elif not isinstance(query, dict):
            raise ValueError('Query parameter must be a dictionary')

        if filters:
            query_filters = query.get("filters", [])
            for key, value in filters.items():
                if isinstance(value, list):
                    query_filters.append({"field": key, "values": value})
                else:
                    query_filters.append({"field": key, "values": [value]})
            query["filters"] = query_filters
        return query

    def _check_identifier(self, id: str) -> str:
        # Perform any preprocessing on the identifier
        if self.asset_type == AssetsConstants.SESSIONS_ASSET_ID:
            id = standardize_name(id)
        return id

    def download_data_with_query(self,
                                 destination_dir: str,
                                 query: Optional[dict] = None,
                                 download_metadata=True) -> dict:
        """Download asset files using a search query.

        Parameters
        ----------
        destination_dir:
            The existing local directory to download the files to.
        query:
            Query to locate the data subset of interest.  Filters through existing
            assets according to the dictionary passed.
        download_metadata:
            If True, downloads asset metadata.  Otherwise, does not download asset metadata.

        Returns
        -------
        dict
            Maps asset IDs obtained with ``query`` to local filepath and metadata.
            Highest level keys represent asset IDs. Lower level keys "filepath" and "metadata" map to local filepath and
            metadata for a specific ID.
        """

        if not os.path.isdir(destination_dir):
            raise NotADirectoryError(
                "{} does not exist".format(destination_dir))

        ids = self.search_assets(
            query=query, scan=True)

        local_dataset = {}

        def threaded_download_asset_and_metadata(id: str):
            logger.info(
                "downloading binary and metadata for {}".format(id))
            local_access = {}
            if not self.metadata_only:
                filepath = os.path.join(destination_dir, id)
                self.get_file_and_write(id=id, write_filepath=filepath)
                local_access["filepath"] = filepath
            if download_metadata:
                metadata = self.get_metadata(id=id)
                local_access["metadata"] = metadata
                json_filepath = os.path.join(
                    destination_dir, '{}.json'.format(id))
                with open(json_filepath, 'w') as file:
                    file.write(json.dumps(metadata))

            local_dataset[id] = local_access

        run_threaded(func=threaded_download_asset_and_metadata,
                     iterable=ids,
                     num_threads=self.num_threads,
                     desc="download data with query")

        return local_dataset

    def get_temporary_upload_url(self,
                                 ids: List[str],
                                 owner_organization_id: Optional[str] = None,
                                 content_type: Optional[str] = None,
                                 ) -> Dict:
        params = self.add_asset_manager_params()
        body = {"ids": ids}

        if owner_organization_id:
            body["owner_organization_id"] = owner_organization_id

        if content_type:
            body["content_type"] = content_type

        resp = requests.post(
            self.get_assets_url('temporary_upload_url/'),
            json=body,
            params=params)
        if not resp.ok:
            raise AssetError(
                'Could not get temporary upload url::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        return resp.json()

    def get_auxiliary_search_fields(self) -> List[dict]:
        all_search_fields = self.get_all_search_fields()
        return [
            d for d in all_search_fields if d['field'].startswith(
                AUXILIARY_KEY + ".")]

    def get_all_search_fields(self) -> List[dict]:
        resp = requests.get(
            self.get_assets_url("meta/search_fields/"),
            params=self.add_asset_manager_params()
        )
        if not resp.ok:
            raise AssetError(
                'Could not get search fields::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        return resp.json()

    def get_url_from_object_access(self,
                                   object_access: ObjectAccess) -> str:

        if object_access.url:
            return object_access.url

        params = self.add_asset_manager_params()
        body = object_access.to_dic()

        resp = requests.post(
            self.get_assets_url('url_from_object_access/'),
            json=body,
            params=params
        )
        if not resp.ok:
            raise AssetError('Could not get url for {}::: {} -- {} -- {}'.format(
                object_access.to_dic(), self.asset_type, resp.reason, resp.content))

        return resp.json()["url"]

    def get_bytes_from_object_access(self,
                                     object_access: ObjectAccess) -> bytes:

        url = self.get_url_from_object_access(
            object_access=object_access)

        resp = requests.get(url)
        if not resp.ok:
            raise AssetError(
                'Could not get url ::: {} -- {} -- {}'.format(
                    self.asset_type, resp.reason, resp.content))
        return resp.content

    def add_auxiliary_metadata(
            self,
            id: Optional[str] = None,
            ids: Optional[List[str]] = None,
            search: Optional[dict] = None,
            filters: Optional[dict] = None,
            **kwargs: Dict[str, Any]
    ):
        """Add auxiliary metadata to an asset.

        Parameters
        ----------
        id:
            An asset IDs to add the auxiliary metadata to.
        ids:
            A list of asset IDs to add the auxiliary metadata to.
        search:
            Query to locate the data subset of interest.  Filters through existing
            assets according to the dictionary passed.
        filters:
            Filter names and values to append to ``search``.  Dict keys represent the existing filter names,
            and dict values are the filter values.
        **kwargs:
            Maps custom metadata labels to any value.  Keys represent the auxiliary metadata labels.
            Values represent the auxiliary metadata values.
        """

        _ids = []
        if id:
            _ids.append(id)

        if ids:
            _ids.extend(ids)

        if search is not None or filters:
            _ids.extend(self.search_assets(query=search, filters=filters, scan=True))
        metadata = {AUXILIARY_KEY: kwargs}

        def _update_metadata_threaded(id_):
            self.update_metadata(id=id_,
                                 metadata=metadata,
                                 replace_sets=False)

        disable_threading = len(_ids) < 2
        run_threaded(func=_update_metadata_threaded,
                     iterable=_ids,
                     num_threads=self.num_threads,
                     disable_threading=disable_threading,
                     disable_tqdm=disable_threading)

    def clear_cache(self, all_organizations: bool = False, partitions: Optional[List[str]]=None):
        '''
        Clears asset manager's Redis cache by organization
        Args:
            all_organizations: if True, Redis cache for all organization is cleared. Requires Super Admin privilege
            partitions: list of partitions (like images, sets, annotations) to clear their cache, if not set, all is cleared

        Returns:
            ACK_OK_RESPONSE on success
        '''

        url = join(self._get_root_url(),'clear_cache')
        params = {"all_organizations": all_organizations}
        if self.auth_token:
            params['token'] = self.auth_token

        payload = {"partitions": partitions} if partitions else {}
        return requests.post(url, params=params,json=payload).json()

    def wait_for_batch_update_task(self,
                                   query_task_id: str,
                                   timeout: int = 1200):
        """Waits for update by query task completion and returns the status."""

        wait_time = 1
        params = {"token": self.auth_token}
        query_task_url = join(self.jobs_namespace_url, f"query_task/{query_task_id}")
        task_status = requests.get(query_task_url, params=params).json()

        start = time.time()
        while task_status.get("completed") is not None and task_status.get("completed") != True:
            time.sleep(wait_time)
            # for small batches, it's too much to wait 5 seconds, start from 1 seconds and incrementally reach to 5
            # don't go over 5 seconds
            wait_time = (wait_time + 1) % 5
            task_status = requests.get(query_task_url, params=params).json()
            duration = start - time.time()
            if duration >= timeout:
                requests.delete(query_task_url, params=params)
                raise TimeoutError(
                    f"Update by query task {query_task_id} taking longer than {timeout} seconds to complete. Task cancelled.")

        return task_status.get("response", {})


"""time tools tools Copyright 2020 Caliber Data Labs."""
#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.
#

import math
import sys
from datetime import datetime, timedelta
from typing import List, Optional, Tuple, Union

import dateutil.parser
import pytz
from dateutil.relativedelta import relativedelta

from ..constants import Time

START_OF_TIME_STRING = Time.START_OF_TIME_STRING
START_OF_TIME_DATETIME = Time.START_OF_TIME_DATETIME
END_OF_TIME_STRING = Time.END_OF_TIME_STRING
END_OF_TIME_DATETIME = Time.END_OF_TIME_DATETIME

epoch = datetime.utcfromtimestamp(0).replace(tzinfo=pytz.utc)


def day_string_from_isoformat_string(ios_timestamp):
    return ios_timestamp.split("T")[0]


def day_string_from_timestamp(timestamp):
    """getting the YYYY-MM-DD time format from a timestamp.

    :param timestamp: input timestamp
    :return: day string in YYYY-MM-DD
    """
    return "{:04d}-{:02d}-{:02d}".format(timestamp.year,
                                         timestamp.month, timestamp.day)


def convert_dict_timestamps_to_string(d, keys):
    for k in keys:
        val = d.get(k)
        if val and isinstance(val, datetime):
            d[k] = datetime_to_iso_utc(val)
    return d


def get_millisecond_from_epoch(timestamp):
    return int((timestamp - epoch).total_seconds() * 1000)


def get_seconds_from_epoch(timestamp):
    return (timestamp - epoch).total_seconds()


def days_between(start_date, end_date):
    end = start_date
    while end < end_date:
        start = end
        end = start + timedelta(1)
        yield start, end


def string_to_datetime(date_string):
    try:
        return dateutil.parser.parse(date_string).astimezone(tz=pytz.utc)
    except ValueError:
        return dateutil.parser.parse(date_string).replace(tzinfo=pytz.utc)


def epoch_to_iso_utc(epoch_time):
    return datetime.utcfromtimestamp(
        epoch_time).replace(tzinfo=pytz.utc).isoformat()


def epoch_to_iso(epoch_time):
    return datetime.utcfromtimestamp(epoch_time).replace(tzinfo=pytz.utc)


def datetime_to_iso_utc(timestamp: datetime) -> str:
    try:
        return timestamp.astimezone(tz=pytz.utc).isoformat()
    except ValueError:
        return timestamp.replace(tzinfo=pytz.utc).isoformat()


def datetime_or_str_to_iso_utc(timestamp: Union[datetime, str]) -> str:
    if isinstance(timestamp, str):
        return timestamp
    else:
        return datetime_to_iso_utc(timestamp)


def str_or_datetime_to_datetime(timestamp: Union[datetime, str]) -> datetime:
    if isinstance(timestamp, datetime):
        return timestamp
    else:
        return string_to_datetime(timestamp)


def datetime_from_seconds(seconds):
    return datetime.fromtimestamp(seconds)


def revert_to_datetime_utc(timestamp):
    """revert string or timestamp to tz_aware timestamp in UTC.

    :param timestamp:
    :return:
    """

    # if datetime
    if isinstance(timestamp, datetime):
        try:
            return timestamp.astimezone(tz=pytz.utc)
        except ValueError:
            return timestamp.replace(tzinfo=pytz.utc)
    # if string
    else:
        return string_to_datetime(timestamp)


def get_relative_time_delta(interval_size, num_intervals, negative=True):
    if negative:
        sign = -1
    else:
        sign = 1

    interval_delta = sign * num_intervals
    time_map = {
        "1h": relativedelta(hours=interval_delta),
        "1d": relativedelta(days=interval_delta),
        "1w": relativedelta(weeks=interval_delta),
        "1M": relativedelta(months=interval_delta),
        "1y": relativedelta(years=interval_delta),
    }
    if interval_size in time_map:
        return time_map[interval_size]
    else:
        raise ValueError("Invalid interval size : {}".format(interval_size))


def get_datetime_utc_from_ros_time(ros_time) -> datetime:
    return datetime.utcfromtimestamp(
        ros_time.secs + ros_time.nsecs * 1e-9).replace(tzinfo=pytz.utc)


def get_time_delta_from_interval(interval: str) -> relativedelta:
    """Get the time delta for the interval."""
    if interval == "second":
        return relativedelta(seconds=1)
    elif interval == "minute":
        return relativedelta(minutes=1)
    elif interval == "hour":
        return relativedelta(hours=1)
    else:
        raise ValueError("Supported intervals are second,minute,hour")


def get_datetime_from_year_and_doy(
        year_doy: str,
        time_zone: str = None) -> datetime:
    if not time_zone:
        return (datetime(int(year_doy[:4]), 1, 1, 12, 0, 0)
                + timedelta(int(year_doy[-3:]) - 1)).astimezone(tz=pytz.utc)
    return (datetime(int(year_doy[:4]), 1, 1, 12, 0, 0)
            + timedelta(int(year_doy[-3:]) - 1)).astimezone(tz=pytz.timezone(time_zone))


def add_minutes(t: datetime, delta_t: int = 60) -> datetime:
    return t + timedelta(minutes=delta_t)


def convert_timestamp_to_time_index(
        resolution: float,
        seconds_since_epoch: float) -> int:
    return int(
        math.floor(
            (seconds_since_epoch
             - Time.START_OF_TIME_DATETIME.timestamp()) /
            resolution))


def merge_time_intervals(
        times: List[Tuple[datetime, datetime]]) -> Tuple[Tuple[datetime, datetime], bool]:
    has_overlap = True
    merged_time = times[0]
    st, et = times[1]
    if st > merged_time[1] or et < merged_time[0]:
        has_overlap = False
        return merged_time, has_overlap
    elif st > merged_time[0] and et < merged_time[1]:
        return merged_time, has_overlap
    elif merged_time[0] < et < merged_time[1]:
        return (st, merged_time[1]), has_overlap
    elif merged_time[0] < st < merged_time[1]:
        return (merged_time[0], et), has_overlap


def time_index_to_timestamp(time_index: int, resolution: float) -> datetime:
    return Time.START_OF_TIME_DATETIME + \
        timedelta(seconds=time_index * resolution)


def get_elapsed_time(end_time: datetime, start_time: datetime):
    return (end_time - start_time).total_seconds()


def jsonify_metadata(metadata: dict,
                     timestamp_keys: Optional[list] = None) -> dict:
    if timestamp_keys is None:
        timestamp_keys = ["timestamp", "start_time", "end_time"]
    # Format the timestamp as str
    for timestamp_key in timestamp_keys:
        timestamp = metadata.get(timestamp_key)
        if timestamp:
            if isinstance(timestamp, datetime):
                metadata[timestamp_key] = timestamp.astimezone(
                    tz=pytz.utc).isoformat()
    return metadata



import os
dir_path = os.path.join(os.path.dirname(os.path.realpath(__file__)))

if os.path.exists(os.path.join(dir_path, "BASE_VERSION")):
    with open(os.path.join(dir_path, "BASE_VERSION")) as version_file:
        __version__ = version_file.readline()
elif os.path.exists(os.path.join(dir_path, "..", "..", "BASE_VERSION")):
    with open(os.path.join(dir_path, "..", "..", "BASE_VERSION")) as version_file:
        __version__ = version_file.readline()

else:
    __version__ = "unknown"


from typing import Optional, Any


class KVStoreTemplate:
    def __init__(self):
        pass

    def put_str(self, key, value):
        pass

    def get_str(self, key, default=None):
        pass

    def put_bytes(self, key, value):
        pass

    def get_bytes(self, key, default=None):
        pass

    def put_json(self, key, json_data):
        pass

    def get_json(self, key: str, default: Optional[Any] = None):
        pass

    def delete(self, key):
        pass

    def clear_all(self, pattern: str):
        pass



"""Miscellaneous (catch all) tools Copyright 2020 Caliber Data Labs."""

#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.
#
import hashlib
import json
import os
import platform
import re
import time
import uuid
import requests
from concurrent.futures import ThreadPoolExecutor
from functools import wraps
from typing import Callable, Dict, Iterable, Optional
from tqdm import tqdm as tqdm_base

from ..custom_exceptions import ResponseError
from ..constants import MAX_THREADS


def decorate_all_inherited_methods(decorator):
    """Add a decorator to all methods which are not private.

    Usage:
        @decorate_all_inherited_methods(decorator)
        class C(object):
            def m1(self): pass
            def m2(self, x): pass
    ...
    """

    def decorate(cls):
        for attr in dir(cls):
            # Only callable methods and those that don't start with _ are
            # eligible
            if callable(getattr(cls, attr)) and not attr.startswith("_"):
                # Only inherited or overrided methods are eligible
                if attr not in cls.__dict__ or hasattr(super(cls), attr):
                    setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls

    return decorate


def chunk_list(list_elements, chunk_size):
    """Chunks a list into chunk_size chunks, with last chunk the remaining
    elements.

    :param list_elements:
    :param chunk_size:
    :return:
    """
    chunks = []
    while True:
        if len(list_elements) > chunk_size:
            chunk = list_elements[0:chunk_size]
            list_elements = list_elements[chunk_size:]
            chunks.append(chunk)
        else:
            chunks.append(list_elements)
            break
    return chunks


def as_bool(s):
    if s is None:
        return False

    if isinstance(s, bool):
        return s
    else:
        s = str(s)  # In Python 2, handle case when s is unicode
        if s.lower() in {"t", "true"}:
            return True
        elif s.lower() in {"f", "false"}:
            return False
        else:
            raise ValueError(
                "Input of type ::: {} cannot be converted to bool ::: {}".format(
                    type(s), s))


def get_truncated_uid(n: int = 5):
    return get_guid()[0:n]


def get_guid():
    return str(uuid.uuid4()).replace('-', '_')


def get_md5_from_file(filename):
    hash_md5 = hashlib.md5()
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def is_valid_hex(color: str) -> bool:

    if re.search(r'^#(?:[0-9a-fA-F]{3}){1,2}$', color):
        return True
    else:
        return False


def hex2rgb(hex: str) -> (int, int, int):
    hex = hex.lstrip('#')
    return tuple(int(hex[i:i + 2], 16) for i in (0, 2, 4))


def rgb2hex(r: int, g: int, b: int) -> str:
    return '#{:02x}{:02x}{:02x}'.format(r, g, b)


def rgb2gray(rgb: tuple) -> int:
    r, g, b = rgb
    return min(round(0.2989 * r + 0.5870 * g + 0.1140 * b), 255)


def get_md5_from_string(string):
    m = hashlib.md5()
    m.update(string.encode("utf-8"))
    return m.hexdigest()


def hash_string_into_positive_integer(string) -> int:
    """return a signed 64-bit."""
    return abs((hash(string) + 2**63) % 2**64 - 2**63)


def hash_string_into_positive_integer_reproducible(string) -> int:
    """return a signed 64-bit."""
    return abs((int(string, 16) + 2**63) % 2**64 - 2**63)


def get_md5_from_json_object(json_object: Dict):
    """get the md5 from a json object (dic, list, etc). We take care of sorting
    the keys here for repeatability.

    :param json_object: list, dic, etc
    :return: md5 of the object
    """
    return get_md5_from_string(json.dumps(json_object, sort_keys=True))


def parse_file_path(input_str: str) \
        -> (str, str, str, str):
    cleaned_str = input_str.replace("\\", "/")
    filename = cleaned_str.split("/")[-1]
    extension = filename.split(".")[-1]
    filename_no_extension = filename.replace("." + extension, "")
    path = cleaned_str.replace(filename, "")
    return filename, path, filename_no_extension, extension


def flatten_json(y: dict, seperator="_") -> dict:
    out = {}

    def flatten(x, name=''):
        if isinstance(x, dict):
            for a in x:
                flatten(x[a], name + a + seperator)
        elif isinstance(x, list):
            i = 0
            for a in x:
                flatten(a, name + str(i) + seperator)
                i += 1
        else:
            out[name[:-1]] = x

    flatten(y)
    return out


def retry(
        exceptions,
        total_tries=4,
        initial_wait=0.5,
        backoff_factor=2,
        logger=None):
    """calling the decorated function applying an exponential backoff.

    Args:
        exceptions: Exception(s) that trigger a retry, can be a tuple
        total_tries: Total tries
        initial_wait: Time to first retry
        backoff_factor: Backoff multiplier (e.g. value of 2 will double the delay each retry).
        logger: logger to be used, if none specified print
    """
    def retry_decorator(f):
        @wraps(f)
        def func_with_retries(*args, **kwargs):
            retry_flag = kwargs.get("retry", False)
            if retry_flag:
                _tries, _delay = total_tries + 1, initial_wait
                while _tries > 1:
                    try:
                        return f(*args, **kwargs)
                    except exceptions as e:
                        logger.info(f'{total_tries + 2 - _tries}. try:')
                        _tries -= 1
                        print_args = args if args else 'no args'
                        if _tries == 1:
                            msg = str(
                                f'Function: {f.__name__}\n'
                                f'Failed despite best efforts after {total_tries} tries.\n'
                                f'args: {print_args}, kwargs: {kwargs}')
                            logger.info(msg)
                            raise
                        msg = str(
                            f'Function: {f.__name__}\n'
                            f'Exception: {e}\n'
                            f'Retrying in {_delay} seconds!, args: {print_args}, kwargs: {kwargs}\n')
                        logger.info(msg)
                        time.sleep(_delay)
                        _delay *= backoff_factor
            else:
                return f(*args, **kwargs)
        return func_with_retries
    return retry_decorator


@retry(exceptions=(Exception, ResponseError), initial_wait=10, backoff_factor=2)
def requests_with_retry(url, payload, headers, type="post"):
    if type == "post":
        resp = requests.post(url,
                             data=payload,
                             headers=headers)
    elif type == "put":
        resp = requests.put(url,
                            data=payload,
                            headers=headers)
    else:
        raise NotImplementedError

    if not resp.ok:
        raise ResponseError("{} ::: {}".format(resp.reason, resp.content))
    else:
        return resp

class tqdm(tqdm_base):
    def __init__(self,
                 progress_callback: Optional[Callable] = None,
                 **kwargs):
        self.progress_callback = progress_callback
        super(tqdm, self).__init__(**kwargs)

    def update(self, n=1):
        super(tqdm, self).update(n)
        if self.total and self.progress_callback:
            progress = float(100.0 * self.n/self.total)
            self.progress_callback(progress)


def run_threaded(func: Callable,
                 iterable: Iterable,
                 num_threads: int = MAX_THREADS,
                 disable_threading: bool = False,
                 disable_tqdm: bool = False,
                 expand_iterable: bool = False,
                 desc: Optional[str] = None,
                 second_iterable: Optional[Iterable] = None,
                 progress_callback: Optional[Callable] = None,
                 **kwargs):
    def _func(_it, _second_it=None):
        if _second_it:
            func(_it, _second_it, **kwargs)
        else:
            func(*_it, **kwargs) if expand_iterable else func(_it, **kwargs)

    if disable_threading is False:
        if second_iterable is None:
            with ThreadPoolExecutor(max_workers=num_threads) as executor:
                list(
                    tqdm(
                        progress_callback=progress_callback,
                        iterable=executor.map(
                            _func,
                            iterable),
                        total=len(
                            list(iterable)),
                        desc=desc,
                        disable=disable_tqdm))
        else:
            assert len(list(iterable)) == len(list(second_iterable))
            with ThreadPoolExecutor(max_workers=num_threads) as executor:
                list(
                    tqdm(
                        progress_callback=progress_callback,
                        iterable=executor.map(
                            _func,
                            iterable,
                            second_iterable),
                        total=len(
                            list(iterable)),
                        desc=desc,
                        disable=disable_tqdm))
    else:
        if second_iterable is None:
            for it in tqdm(progress_callback=progress_callback, iterable=iterable, desc=desc, disable=disable_tqdm):
                _func(it)
        else:
            raise NotImplementedError("using second iterable in non threaded mode is not supported")


def join(*args):
    if platform.system() == "Windows":
        return os.path.join(*args).replace("\\", "/")
    else:
        return os.path.join(*args)


#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.
#

from typing import Optional

from shared.external.constants import (AnnotationMediaTypes,
                                       AnnotationProviders, AnnotationTypes, AssetsConstants, AnnotationGroups)
from shared.external.custom_exceptions import InvalidAnnotationError
from shared.external.models.annotation import Annotation, AnnotationEntity
from shared.external.models.object_access import ObjectAccess
from shared.external.tools import time_utils
from shared.external.tools.misc import get_md5_from_string


class ScaleAnnotation(Annotation):
    def __init__(
            self,
            scale_response: dict,
            version: Optional[str] = None):

        self.scale_response = scale_response
        self.scaleapi_task = scale_response.get('task', {})
        self.scaleapi_params = self.scaleapi_task.get('params', {})
        self.scaleapi_type = self.scaleapi_task.get('type')
        self.metadata = self.scaleapi_task.get('metadata', {})
        self.scaleapi_annotations = scale_response.get(
            'response', {}).get('annotations', {})

        if self.scaleapi_type == 'lidarannotation':
            media_type = AssetsConstants.LIDARS_ASSET_ID
        else:
            media_type = AssetsConstants.IMAGES_ASSET_ID

        self.sets = [self.metadata.get("output_set_id")]
        if self.scaleapi_type == 'annotation':
            annotation_type = AnnotationTypes.TWO_D_BOUNDING_BOX
        elif self.scaleapi_type == 'segmentannotation':
            annotation_type = AnnotationTypes.SEGMENTATION
        elif self.scaleapi_type == 'polygonannotation':
            annotation_type = AnnotationTypes.POLYGON
        elif self.scaleapi_type == 'imageannotation':
            annotation_type = AnnotationTypes.LINE
        elif self.scaleapi_type == 'lidarannotation':
            annotation_type = AnnotationTypes.CUBOID
        else:
            raise InvalidAnnotationError(
                'invalid annotation type {}'.format(
                    self.scaleapi_type))

        super().__init__(
            annotation_meta=scale_response,
            asset_id=self.metadata.get("asset_id"),
            annotation_type=annotation_type,
            media_type=media_type,
            provider=AnnotationProviders.SCALE,
            version=version,
            annotation_group=AnnotationGroups.GROUND_TRUTH,
            id=scale_response.get('task_id') + ".ann",
            timestamp=time_utils.string_to_datetime(self.scaleapi_task.get('created_at')),
            set_id=self.metadata.get("output_set_id")
        )

        self.scaleapi_annotations = scale_response.get(
            'response', {}).get('annotations', {})

        if self.scaleapi_task.get('status') != 'completed':
            raise InvalidAnnotationError(
                'status {} is not completed'.format(
                    self.scaleapi_task.get('status')))

        self.finalize()

    def set_annotation_entities(self):
        if self.scaleapi_type == 'annotation':
            self.__set_bounding_box_annotation_entities()
        elif self.scaleapi_type == 'segmentannotation':
            self.__set_segmentation_annotation_entities()
        elif self.scaleapi_type == 'polygonannotation':
            self.__set_polygon_annotation_entities()
        elif self.scaleapi_type == 'imageannotation':
            self.__set_line_annotation_entities()
        elif self.scaleapi_type == 'lidarannotation':
            self.__set_lidar_annotation_entities()
        else:
            raise InvalidAnnotationError(
                'invalid annotation type {}'.format(
                    self.scaleapi_type))

    def __set_bounding_box_annotation_entities(self):

        self.annotation_entities = []

        if not isinstance(self.scaleapi_annotations, list):
            raise InvalidAnnotationError(
                'bounding box annotation is expected to be list')

        for scaleapi_annotation in self.scaleapi_annotations:
            try:
                width = scaleapi_annotation['width']
                height = scaleapi_annotation['height']
                left = scaleapi_annotation['left']
                top = scaleapi_annotation['top']
                label = scaleapi_annotation['label']
            except KeyError as e:
                raise InvalidAnnotationError(str(e))

            coordinates = [
                {
                    'x': left,
                    'y': top
                },
                {
                    'x': left + width,
                    'y': top
                },
                {
                    'x': left + width,
                    'y': top + height
                },
                {
                    'x': left,
                    'y': top + height
                }
            ]
            self.annotation_entities.append(AnnotationEntity(
                label=label,
                annotation_type=AnnotationTypes.TWO_D_BOUNDING_BOX,
                coordinates=coordinates,
                attributes=scaleapi_annotation.get('attributes')
            ))

    def __set_segmentation_annotation_entities(self):

        if not isinstance(self.scaleapi_annotations, dict):
            raise InvalidAnnotationError(
                'bounding box annotation is expected to be dict')

        self.annotation_entities = []

        if 'unlabeled' in self.scaleapi_annotations:
            url = self.scaleapi_annotations['unlabeled']
            mask_id = str(get_md5_from_string(url)).replace('-', '_')
            self.masks[mask_id] = ObjectAccess(url=url)
            self.annotation_entities.append(
                # TODO mask color needs to be specified here
                AnnotationEntity(
                    label='unlabeled',
                    annotation_type=AnnotationTypes.SEGMENTATION,
                    mask_id=mask_id,
                    mask_color="#ffffff"
                )
            )

        labeled_dic = self.scaleapi_annotations.get('labeled', {})
        for label in labeled_dic:
            value = labeled_dic[label]
            if isinstance(value, list):
                for url in value:
                    mask_id = str(get_md5_from_string(url)).replace('-', '_')
                    self.masks[mask_id] = ObjectAccess(url=url)
                    # TODO mask color needs to be specified here
                    self.annotation_entities.append(AnnotationEntity(
                        label=label,
                        annotation_type=AnnotationTypes.SEGMENTATION,
                        mask_id=mask_id,
                        mask_color="#ffffff")
                    )
            else:
                mask_id = str(get_md5_from_string(value)).replace('-', '_')
                self.masks[mask_id]= ObjectAccess(url=value)
                # TODO mask color needs to be specified here
                self.annotation_entities.append(AnnotationEntity(
                    label=label,
                    annotation_type=AnnotationTypes.SEGMENTATION,
                    mask_id=mask_id,
                    mask_color="#ffffff")
                )

    def __set_polygon_annotation_entities(self):
        self.annotation_entities = []

        if not isinstance(self.scaleapi_annotations, list):
            raise InvalidAnnotationError(
                'polygon annotation is expected to be list')

        for scaleapi_annotation in self.scaleapi_annotations:
            try:
                coordinates = scaleapi_annotation['vertices']
                key = scaleapi_annotation.get('key')
                label = scaleapi_annotation['label']
            except KeyError as e:
                raise InvalidAnnotationError(str(e))

            self.annotation_entities.append(AnnotationEntity(
                label=label,
                annotation_type=AnnotationTypes.POLYGON,
                coordinates=coordinates,
                attributes=scaleapi_annotation.get('attributes'),
                uid=key
            ))

    def __set_line_annotation_entities(self):
        self.annotation_entities = []

        if not isinstance(self.scaleapi_annotations, list):
            raise InvalidAnnotationError(
                'line annotation is expected to be list')

        for scaleapi_annotation in self.scaleapi_annotations:
            try:
                coordinates = scaleapi_annotation['vertices']
                key = scaleapi_annotation['key']
                label = scaleapi_annotation['label']
            except KeyError as e:
                raise InvalidAnnotationError(str(e))

            self.annotation_entities.append(AnnotationEntity(
                label=label,
                annotation_type=AnnotationTypes.LINE,
                coordinates=coordinates,
                attributes=scaleapi_annotation.get('attributes'),
                uid=key
            ))

    def __set_lidar_annotation_entities(self):

        self.annotation_entities = []

        if not isinstance(self.scaleapi_annotations, list):
            raise InvalidAnnotationError(
                'lidar annotation is expected to be list')

        for scaleapi_annotation in self.scaleapi_annotations:
            try:
                width = scaleapi_annotation['width']
                height = scaleapi_annotation['height']
                left = scaleapi_annotation['left']
                top = scaleapi_annotation['top']
                label = scaleapi_annotation['label']
            except KeyError as e:
                raise InvalidAnnotationError(str(e))

            coordinates = [
                {
                    'x': left,
                    'y': top
                },
                {
                    'x': left + width,
                    'y': top
                },
                {
                    'x': left + width,
                    'y': top + height
                },
                {
                    'x': left,
                    'y': top + height
                }
            ]
            self.annotation_entities.append(AnnotationEntity(
                label=label,
                annotation_type=AnnotationTypes.CUBOID,
                coordinates=coordinates,
                attributes=scaleapi_annotation.get('attributes')
            ))



#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.
#
import os

import mock
from shared.external.models.annotation_transformers.scale import \
    ScaleAnnotation
from shared.tools.misc import json_file_to_dic

my_path = path = os.path.dirname(os.path.abspath(__file__))


def test_get_scaleapi():
    annotation = ScaleAnnotation(json_file_to_dic(
        os.path.join(my_path, 'resources/889fa13h29.ann')))
    assert 'polygon' == annotation.annotation_type
    assert '2019-05-14-17-38-55left_raw_raw_00362.png' == annotation.asset_id
    assert 'Drivable Space' == annotation.annotation_entities[0].label
    assert 15 == len(annotation.annotation_entities[1].coordinates)
    assert 3 == len(annotation.annotation_entities)


def test_get_scaleapi_2():
    annotation = ScaleAnnotation(json_file_to_dic(
        os.path.join(my_path, 'resources/5ecc2f9736e6f30047b66f58.ann')))
    assert 'polygon' == annotation.annotation_type
    assert 'kitti_2011_09_28_drive_2_kitti_camera_color_right_image_raw-1317213980051.png' == annotation.asset_id
    assert 'Bicyclist' == annotation.annotation_entities[0].label
    assert 3 == len(annotation.annotation_entities[0].coordinates)
    assert ["Bicyclist"] == annotation.to_dic().get("labels")


def test_get_scaleapi_3():
    annotation = ScaleAnnotation(json_file_to_dic(
        os.path.join(my_path, 'resources/5eface514847b300269ab604.ann')))
    assert 'polygon' == annotation.annotation_type
    assert 'axis_front_image_raw_compressed_data_1581036613006.jpeg' == annotation.asset_id
    assert 'Sedan' == annotation.annotation_entities[0].label
    assert 3 == len(annotation.annotation_entities[0].coordinates)
    assert 1 == len(annotation.annotation_entities)
    assert "Sedan" in annotation.to_dic().get("labels")



#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.
#


from typing import List, Optional, Union

from shared.external.constants import AnnotationProviders, AnnotationTypes, AssetsConstants, AnnotationGroups
from shared.external.custom_exceptions import InvalidAnnotationError
from shared.external.models.annotation import Annotation, AnnotationEntity
from shared.external.tools import time_utils


class CocoAnnotation(Annotation):
    def __init__(
        self,
        image_dict: dict,
        annotations_dict: List[dict],
        categories_dict: List[dict],
        annotation_type: AnnotationTypes,
        set_name: Optional[str] = None,
        provider: Optional[Union[AnnotationProviders, str]] = None,
        version: Optional[str] = None,
        annotation_group=AnnotationGroups.GROUND_TRUTH
    ):
        metadata = {"image": image_dict, "categories": categories_dict}
        if annotation_type not in [
            AnnotationTypes.SEGMENTATION,
            AnnotationTypes.TWO_D_BOUNDING_BOX,
            AnnotationTypes.POLYGON,
        ]:
            raise InvalidAnnotationError(
                f"invalid annotation type {annotation_type}"
            )
        if "date_captured" in image_dict:
            timestamp = time_utils.string_to_datetime(
                image_dict["date_captured"]
            )
        else:
            timestamp = None

        asset_id = image_dict.get("file_name")

        super().__init__(
            annotation_meta=metadata,
            set_id=set_name,
            annotation_type=annotation_type,
            annotation_group=annotation_group,
            version=version,
            media_type=AssetsConstants.IMAGES_ASSET_ID,
            timestamp=timestamp,
            provider=provider,
            asset_id=asset_id,
            id=asset_id + "_" + provider
        )

        self.image_dict: dict = image_dict
        self.annotations_dict: List[dict] = annotations_dict
        self.categories_dict: List[dict] = categories_dict

        self.finalize()

    def set_annotation_entities(self):
        if self.annotation_type == AnnotationTypes.TWO_D_BOUNDING_BOX:
            self.__set_bounding_box_annotation_entities()
        elif self.annotation_type == AnnotationTypes.POLYGON:
            self.__set_polygon_annotation_entities()
        else:
            raise InvalidAnnotationError(
                f"invalid annotation type {self.annotation_type}"
            )

    def __set_bounding_box_annotation_entities(self):

        self.annotation_entities = []

        if not isinstance(self.annotations_dict, list):
            raise InvalidAnnotationError(
                "bounding box annotation is expected to be list"
            )

        for annotation in self.annotations_dict:
            try:
                category_id = annotation["category_id"]
                left = float(annotation["bbox"][0])
                top = float(annotation["bbox"][1])
                width = float(annotation["bbox"][2])
                height = float(annotation["bbox"][3])
            except KeyError as e:
                raise InvalidAnnotationError(str(e))

            coordinates = [
                {"x": left, "y": top},
                {"x": left + width, "y": top},
                {"x": left + width, "y": top + height},
                {"x": left, "y": top + height},
            ]
            label = next(
                filter(lambda x: x["id"] == category_id, self.categories_dict)
            )["name"]
            self.annotation_entities.append(
                AnnotationEntity(
                    label=label,
                    category_id=category_id,
                    annotation_type=AnnotationTypes.TWO_D_BOUNDING_BOX,
                    coordinates=coordinates
                )
            )

    def __set_polygon_annotation_entities(self):
        if not isinstance(self.annotations_dict, list):
            raise InvalidAnnotationError(
                "Polygon annotations are expected to be a list"
            )
        self.annotation_entities = []
        for annotation in self.annotations_dict:
            category_id = annotation["category_id"]
            for coordinate_list in annotation["segmentation"]:
                coordinates = []
                iter_list = iter(coordinate_list)
                for x in iter_list:
                    y = next(iter_list)
                    coordinates.append({"x" : x, "y": y})
                label = next(
                    filter(lambda x: x["id"] == category_id, self.categories_dict)
                )["name"]
                self.annotation_entities.append(
                    AnnotationEntity(
                        label=label,
                        category_id=category_id,
                        coordinates=coordinates,
                        annotation_type=AnnotationTypes.POLYGON,
                    )
                )





#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.

from .asset_manager_client import AssetManagerClient
from .scene_engine_client import SceneEngineClient
from .job_manager_client import JobManagerClient



"""custom exceptions for the data platform Copyright 2020 Caliber Data Labs."""
#  Copyright (c) 2020 Caliber Data Labs.
#  All rights reserved.
#


class SearchError(Exception):
    pass


class InvalidFileError(Exception):
    pass


class FileSizeCheckError(Exception):
    pass


class AssetError(Exception):
    pass


class JobError(Exception):
    pass


class CredentialsError(Exception):
    pass


class ResponseError(Exception):
    pass


class SessionManagerError(Exception):
    pass


class InvalidURLError(Exception):
    pass


class IndexingException(Exception):
    pass


class InvalidArgumentsError(Exception):
    pass


class MetadataNotFoundError(Exception):
    pass


class InvalidAnnotationError(Exception):
    pass


class InvalidConfigError(Exception):
    pass


class ZippingError(Exception):
    pass


class SetsError(Exception):
    pass


class InvalidEventError(Exception):
    pass


class DataEnrichmentError(Exception):
    pass


class SessionError(Exception):
    pass


class InvalidTimeError(Exception):
    pass


class ProjectError(Exception):
    pass


class OperationError(Exception):
    pass


class InvalidAuthorization(Exception):
    pass


class CacheError(Exception):
    pass


class UserManagerError(Exception):
    pass


class UserAlreadyExistsError(Exception):
    pass


class InvalidOrganization(Exception):
    pass


class OrganizationAlreadyExistsError(Exception):
    pass


class OwnershipAlreadyExistsError(Exception):
    pass


class IdentityError(Exception):
    pass


class AssetCompressionError(Exception):
    pass


class ThumbnailNotAvailableError(Exception):
    pass


class InvalidInfluenceTypeError(Exception):
    pass


class KafkaProducerError(Exception):
    pass


class KafkaConsumerError(Exception):
    pass


class UsageTrackerError(Exception):
    pass


class QuotaSystemError(Exception):
    pass


class QuotaExceedError(Exception):
    pass


class UploadError(Exception):
    pass


class SimilaritySearchError(Exception):
    pass


class EmbeddingError(Exception):
    pass


class AccessError(Exception):
    pass


class StorageError(Exception):
    pass

